#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Models representing steps runs."""

from datetime import datetime, timedelta
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Dict,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from zenml.config.step_configurations import Step, StepConfiguration, StepSpec
from zenml.constants import STR_FIELD_MAX_LENGTH, TEXT_FIELD_MAX_LENGTH
from zenml.enums import (
    ArtifactSaveType,
    ExecutionStatus,
    StepRunInputArtifactType,
    StepType,
)
from zenml.metadata.metadata_types import MetadataType
from zenml.models.v2.base.base import BaseUpdate
from zenml.models.v2.base.scoped import (
    ProjectScopedFilter,
    ProjectScopedRequest,
    ProjectScopedResponse,
    ProjectScopedResponseBody,
    ProjectScopedResponseMetadata,
    ProjectScopedResponseResources,
    RunMetadataFilterMixin,
)
from zenml.models.v2.core.artifact_version import ArtifactVersionResponse
from zenml.models.v2.core.model_version import ModelVersionResponse
from zenml.models.v2.misc.exception_info import ExceptionInfo
from zenml.utils.time_utils import utc_now

if TYPE_CHECKING:
    from sqlalchemy.sql.elements import ColumnElement

    from zenml.models.v2.core.logs import (
        LogsRequest,
        LogsResponse,
    )
    from zenml.zen_stores.schemas import BaseSchema

    AnySchema = TypeVar("AnySchema", bound=BaseSchema)


class StepRunInputResponse(ArtifactVersionResponse):
    """Response model for step run inputs."""

    input_type: StepRunInputArtifactType
    index: Optional[int] = Field(
        title="The index of the input artifact in the step run.",
        default=None,
    )
    chunk_index: Optional[int] = Field(
        title="The index of the chunk in the input artifact.",
        default=None,
    )
    chunk_size: Optional[int] = Field(
        title="The size of the chunk in the input artifact.",
        default=None,
    )

    def get_hydrated_version(self) -> "StepRunInputResponse":
        """Get the hydrated version of this step run input.

        Returns:
            an instance of the same entity with the metadata field attached.
        """
        from zenml.client import Client

        return StepRunInputResponse(
            input_type=self.input_type,
            **Client().zen_store.get_artifact_version(self.id).model_dump(),
        )


# ------------------ Request Model ------------------


class StepRunRequest(ProjectScopedRequest):
    """Request model for step runs."""

    name: str = Field(
        title="The name of the pipeline run step.",
        max_length=STR_FIELD_MAX_LENGTH,
    )
    start_time: datetime = Field(
        title="The start time of the step run.",
    )
    end_time: Optional[datetime] = Field(
        title="The end time of the step run.",
        default=None,
    )
    status: ExecutionStatus = Field(title="The status of the step.")
    cache_key: Optional[str] = Field(
        title="The cache key of the step run.",
        default=None,
        max_length=STR_FIELD_MAX_LENGTH,
    )
    cache_expires_at: Optional[datetime] = Field(
        title="The time at which this step run should not be used for cached "
        "results anymore. If not set, the result will never expire.",
        default=None,
    )
    code_hash: Optional[str] = Field(
        title="The code hash of the step run.",
        default=None,
        max_length=STR_FIELD_MAX_LENGTH,
    )
    docstring: Optional[str] = Field(
        title="The docstring of the step function or class.",
        default=None,
        max_length=TEXT_FIELD_MAX_LENGTH,
    )
    source_code: Optional[str] = Field(
        title="The source code of the step function or class.",
        default=None,
        max_length=TEXT_FIELD_MAX_LENGTH,
    )
    pipeline_run_id: UUID = Field(
        title="The ID of the pipeline run that this step run belongs to.",
    )
    original_step_run_id: Optional[UUID] = Field(
        title="The ID of the original step run if this step was cached.",
        default=None,
    )
    parent_step_ids: List[UUID] = Field(
        title="The IDs of the parent steps of this step run.",
        default_factory=list,
        deprecated=True,
    )
    inputs: Dict[str, List[UUID]] = Field(
        title="The IDs of the input artifact versions of the step run.",
        default_factory=dict,
    )
    outputs: Dict[str, List[UUID]] = Field(
        title="The IDs of the output artifact versions of the step run.",
        default_factory=dict,
    )
    logs: Optional[Union[UUID, "LogsRequest"]] = Field(
        title="Logs associated with this step run.",
        default=None,
    )
    exception_info: Optional[ExceptionInfo] = Field(
        default=None,
        title="The exception information of the step run.",
    )
    dynamic_config: Optional["Step"] = Field(
        title="The dynamic configuration of the step run.",
        default=None,
    )

    model_config = ConfigDict(protected_namespaces=())


# ------------------ Update Model ------------------


class StepRunUpdate(BaseUpdate):
    """Update model for step runs."""

    outputs: Dict[str, List[UUID]] = Field(
        title="The IDs of the output artifact versions of the step run.",
        default={},
    )
    loaded_artifact_versions: Dict[str, UUID] = Field(
        title="The IDs of artifact versions that were loaded by this step run.",
        default={},
    )
    status: Optional[ExecutionStatus] = Field(
        title="The status of the step.",
        default=None,
    )
    end_time: Optional[datetime] = Field(
        title="The end time of the step run.",
        default=None,
    )
    exception_info: Optional[ExceptionInfo] = Field(
        default=None,
        title="The exception information of the step run.",
    )
    cache_expires_at: Optional[datetime] = Field(
        title="The time at which this step run should not be used for cached "
        "results anymore.",
        default=None,
    )
    add_logs: Optional[List["LogsRequest"]] = Field(
        default=None, title="New logs to add to the step run."
    )
    model_config = ConfigDict(protected_namespaces=())


# ------------------ Response Model ------------------
class StepRunResponseBody(ProjectScopedResponseBody):
    """Response body for step runs."""

    type: Optional[StepType] = Field(
        title="The type of the step.",
        default=None,
    )
    status: ExecutionStatus = Field(title="The status of the step.")
    version: int = Field(
        title="The version of the step run.",
    )
    is_retriable: bool = Field(
        title="Whether the step run is retriable.",
    )
    start_time: Optional[datetime] = Field(
        title="The start time of the step run.",
        default=None,
    )
    end_time: Optional[datetime] = Field(
        title="The end time of the step run.",
        default=None,
    )
    latest_heartbeat: Optional[datetime] = Field(
        title="The latest heartbeat of the step run.",
        default=None,
    )
    model_version_id: Optional[UUID] = Field(
        title="The ID of the model version that was "
        "configured by this step run explicitly.",
        default=None,
    )
    substitutions: Dict[str, str] = Field(
        title="The substitutions of the step run.",
        default={},
    )
    heartbeat_threshold: Optional[int] = Field(
        title="The applied heartbeat healthiness threshold ",
        default=None,
    )
    model_config = ConfigDict(protected_namespaces=())


class StepRunResponseMetadata(ProjectScopedResponseMetadata):
    """Response metadata for step runs."""

    __zenml_skip_dehydration__: ClassVar[List[str]] = [
        "config",
        "spec",
        "metadata",
    ]

    # Configuration
    config: "StepConfiguration" = Field(title="The configuration of the step.")
    spec: "StepSpec" = Field(title="The spec of the step.")

    # Code related fields
    cache_key: Optional[str] = Field(
        title="The cache key of the step run.",
        default=None,
        max_length=STR_FIELD_MAX_LENGTH,
    )
    cache_expires_at: Optional[datetime] = Field(
        title="The time at which this step run should not be used for cached "
        "results anymore. If not set, the result will never expire.",
        default=None,
    )
    code_hash: Optional[str] = Field(
        title="The code hash of the step run.",
        default=None,
        max_length=STR_FIELD_MAX_LENGTH,
    )
    docstring: Optional[str] = Field(
        title="The docstring of the step function or class.",
        default=None,
        max_length=TEXT_FIELD_MAX_LENGTH,
    )
    source_code: Optional[str] = Field(
        title="The source code of the step function or class.",
        default=None,
        max_length=TEXT_FIELD_MAX_LENGTH,
    )
    exception_info: Optional[ExceptionInfo] = Field(
        default=None,
        title="The exception information of the step run.",
    )

    # References
    snapshot_id: UUID = Field(
        title="The snapshot associated with the step run."
    )
    pipeline_run_id: UUID = Field(
        title="The ID of the pipeline run that this step run belongs to.",
    )
    original_step_run_id: Optional[UUID] = Field(
        title="The ID of the original step run if this step was cached.",
        default=None,
    )
    parent_step_ids: List[UUID] = Field(
        title="The IDs of the parent steps of this step run.",
        default_factory=list,
    )
    run_metadata: Dict[str, MetadataType] = Field(
        title="Metadata associated with this step run.",
        default={},
    )


class StepRunResponseResources(ProjectScopedResponseResources):
    """Class for all resource models associated with the step run entity."""

    log_collection: Optional[List["LogsResponse"]] = Field(
        title="Logs associated with this step run.",
        default=None,
    )

    model_version: Optional[ModelVersionResponse] = None
    inputs: Dict[str, List[StepRunInputResponse]] = Field(
        title="The input artifact versions of the step run.",
        default_factory=dict,
    )
    outputs: Dict[str, List[ArtifactVersionResponse]] = Field(
        title="The output artifact versions of the step run.",
        default_factory=dict,
    )

    # TODO: In Pydantic v2, the `model_` is a protected namespaces for all
    #  fields defined under base models. If not handled, this raises a warning.
    #  It is possible to suppress this warning message with the following
    #  configuration, however the ultimate solution is to rename these fields.
    #  Even though they do not cause any problems right now, if we are not
    #  careful we might overwrite some fields protected by pydantic.
    model_config = ConfigDict(protected_namespaces=())


class StepRunResponse(
    ProjectScopedResponse[
        StepRunResponseBody, StepRunResponseMetadata, StepRunResponseResources
    ]
):
    """Response model for step runs."""

    name: str = Field(
        title="The name of the pipeline run step.",
        max_length=STR_FIELD_MAX_LENGTH,
    )

    def get_hydrated_version(self) -> "StepRunResponse":
        """Get the hydrated version of this step run.

        Returns:
            an instance of the same entity with the metadata field attached.
        """
        from zenml.client import Client

        return Client().zen_store.get_run_step(self.id)

    # Helper properties
    @property
    def input(self) -> StepRunInputResponse:
        """Returns the input artifact that was used to run this step.

        Returns:
            The input artifact.

        Raises:
            ValueError: If there were zero or multiple inputs to this step.
        """
        if not self.inputs:
            raise ValueError(f"Step {self.name} has no inputs.")
        if len(self.inputs) > 1 or (
            len(self.inputs) == 1 and len(next(iter(self.inputs.values()))) > 1
        ):
            raise ValueError(
                f"Step {self.name} has multiple inputs, so `Step.input` is "
                "ambiguous. Please use `Step.inputs` instead."
            )
        return next(iter(self.inputs.values()))[0]

    @property
    def output(self) -> ArtifactVersionResponse:
        """Returns the output artifact that was written by this step.

        Returns:
            The output artifact.

        Raises:
            ValueError: If there were zero or multiple step outputs.
        """
        if not self.outputs:
            raise ValueError(f"Step {self.name} has no outputs.")
        if len(self.outputs) > 1 or (
            len(self.outputs) == 1
            and len(next(iter(self.outputs.values()))) > 1
        ):
            raise ValueError(
                f"Step {self.name} has multiple outputs, so `Step.output` is "
                "ambiguous. Please use `Step.outputs` instead."
            )
        return next(iter(self.outputs.values()))[0]

    @property
    def regular_inputs(self) -> Dict[str, List[StepRunInputResponse]]:
        """Returns the regular step inputs of the step run.

        Regular step inputs are the inputs that are defined in the step function
        signature, and are not manually loaded during the step execution.

        Returns:
            The regular step inputs.
        """
        result = {}

        for input_name, input_artifacts in self.inputs.items():
            filtered = [
                input_artifact
                for input_artifact in input_artifacts
                if input_artifact.input_type != StepRunInputArtifactType.MANUAL
            ]
            if filtered:
                result[input_name] = filtered

        return result

    @property
    def regular_outputs(self) -> Dict[str, ArtifactVersionResponse]:
        """Returns the regular step outputs of the step run.

        Regular step outputs are the outputs that are defined in the step
        function signature, and are not manually saved during the step
        execution.

        Raises:
            ValueError: If there were multiple regular output artifacts for the
                same output name.

        Returns:
            The regular step outputs.
        """
        result = {}

        for output_name, output_artifacts in self.outputs.items():
            filtered = [
                output_artifact
                for output_artifact in output_artifacts
                if output_artifact.save_type == ArtifactSaveType.STEP_OUTPUT
            ]
            if len(filtered) > 1:
                raise ValueError(
                    f"Expected 1 regular output artifact for {output_name}, "
                    f"got {len(filtered)}."
                )
            if filtered:
                result[output_name] = filtered[0]

        return result

    # Body and metadata properties

    @property
    def type(self) -> Optional[StepType]:
        """The `type` property.

        Returns:
            the value of the property.
        """
        return self.get_body().type

    @property
    def status(self) -> ExecutionStatus:
        """The `status` property.

        Returns:
            the value of the property.
        """
        return self.get_body().status

    @property
    def version(self) -> int:
        """The `version` property.

        Returns:
            the value of the property.
        """
        return self.get_body().version

    @property
    def is_retriable(self) -> bool:
        """The `is_retriable` property.

        Returns:
            the value of the property.
        """
        return self.get_body().is_retriable

    @property
    def inputs(self) -> Dict[str, List[StepRunInputResponse]]:
        """The `inputs` property.

        Returns:
            the value of the property.
        """
        return self.get_resources().inputs

    @property
    def outputs(self) -> Dict[str, List[ArtifactVersionResponse]]:
        """The `outputs` property.

        Returns:
            the value of the property.
        """
        return self.get_resources().outputs

    @property
    def model_version_id(self) -> Optional[UUID]:
        """The `model_version_id` property.

        Returns:
            the value of the property.
        """
        return self.get_body().model_version_id

    @property
    def substitutions(self) -> Dict[str, str]:
        """The `substitutions` property.

        Returns:
            the value of the property.
        """
        return self.get_body().substitutions

    @property
    def config(self) -> "StepConfiguration":
        """The `config` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().config

    @property
    def spec(self) -> "StepSpec":
        """The `spec` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().spec

    @property
    def cache_key(self) -> Optional[str]:
        """The `cache_key` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().cache_key

    @property
    def cache_expires_at(self) -> Optional[datetime]:
        """The `cache_expires_at` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().cache_expires_at

    @property
    def code_hash(self) -> Optional[str]:
        """The `code_hash` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().code_hash

    @property
    def docstring(self) -> Optional[str]:
        """The `docstring` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().docstring

    @property
    def source_code(self) -> Optional[str]:
        """The `source_code` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().source_code

    @property
    def start_time(self) -> Optional[datetime]:
        """The `start_time` property.

        Returns:
            the value of the property.
        """
        return self.get_body().start_time

    @property
    def end_time(self) -> Optional[datetime]:
        """The `end_time` property.

        Returns:
            the value of the property.
        """
        return self.get_body().end_time

    @property
    def duration(self) -> Optional[timedelta]:
        """The `duration` property.

        Returns:
            The duration of the step run.
        """
        if self.end_time is None or self.start_time is None:
            return None
        return self.end_time - self.start_time

    @property
    def latest_heartbeat(self) -> Optional[datetime]:
        """The `latest_heartbeat` property.

        Returns:
            the value of the property.
        """
        return self.get_body().latest_heartbeat

    @property
    def heartbeat_threshold(self) -> Optional[int]:
        """The `heartbeat_threshold` property.

        Returns:
            the value of the property.
        """
        return self.get_body().heartbeat_threshold

    @property
    def snapshot_id(self) -> UUID:
        """The `snapshot_id` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().snapshot_id

    @property
    def pipeline_run_id(self) -> UUID:
        """The `pipeline_run_id` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().pipeline_run_id

    @property
    def original_step_run_id(self) -> Optional[UUID]:
        """The `original_step_run_id` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().original_step_run_id

    @property
    def parent_step_ids(self) -> List[UUID]:
        """The `parent_step_ids` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().parent_step_ids

    @property
    def exception_info(self) -> Optional[ExceptionInfo]:
        """The `exception_info` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().exception_info

    @property
    def run_metadata(self) -> Dict[str, MetadataType]:
        """The `run_metadata` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().run_metadata

    @property
    def log_collection(self) -> Optional[List["LogsResponse"]]:
        """The `log_collection` property.

        Returns:
            the value of the property.
        """
        return self.get_resources().log_collection

    @property
    def model_version(self) -> Optional[ModelVersionResponse]:
        """The `model_version` property.

        Returns:
            the value of the property.
        """
        return self.get_resources().model_version


# ------------------ Filter Model ------------------


class StepRunFilter(ProjectScopedFilter, RunMetadataFilterMixin):
    """Model to enable advanced filtering of step runs."""

    FILTER_EXCLUDE_FIELDS: ClassVar[List[str]] = [
        *ProjectScopedFilter.FILTER_EXCLUDE_FIELDS,
        *RunMetadataFilterMixin.FILTER_EXCLUDE_FIELDS,
        "model",
        "exclude_retried",
        "cache_expired",
    ]
    CLI_EXCLUDE_FIELDS: ClassVar[List[str]] = [
        *ProjectScopedFilter.CLI_EXCLUDE_FIELDS,
        *RunMetadataFilterMixin.CLI_EXCLUDE_FIELDS,
    ]
    CUSTOM_SORTING_OPTIONS: ClassVar[List[str]] = [
        *ProjectScopedFilter.CUSTOM_SORTING_OPTIONS,
        *RunMetadataFilterMixin.CUSTOM_SORTING_OPTIONS,
    ]
    API_MULTI_INPUT_PARAMS: ClassVar[List[str]] = [
        *ProjectScopedFilter.API_MULTI_INPUT_PARAMS,
        *RunMetadataFilterMixin.API_MULTI_INPUT_PARAMS,
    ]

    name: Optional[str] = Field(
        default=None,
        description="Name of the step run",
    )
    code_hash: Optional[str] = Field(
        default=None,
        description="Code hash for this step run",
    )
    cache_key: Optional[str] = Field(
        default=None,
        description="Cache key for this step run",
    )
    status: Optional[str] = Field(
        default=None,
        description="Status of the Step Run",
    )
    start_time: Optional[Union[datetime, str]] = Field(
        default=None,
        description="Start time for this run",
        union_mode="left_to_right",
    )
    end_time: Optional[Union[datetime, str]] = Field(
        default=None,
        description="End time for this run",
        union_mode="left_to_right",
    )
    pipeline_run_id: Optional[Union[UUID, str]] = Field(
        default=None,
        description="Pipeline run of this step run",
        union_mode="left_to_right",
    )
    snapshot_id: Optional[Union[UUID, str]] = Field(
        default=None,
        description="Snapshot of this step run",
        union_mode="left_to_right",
    )
    original_step_run_id: Optional[Union[UUID, str]] = Field(
        default=None,
        description="Original id for this step run",
        union_mode="left_to_right",
    )
    model_version_id: Optional[Union[UUID, str]] = Field(
        default=None,
        description="Model version associated with the step run.",
        union_mode="left_to_right",
    )
    model: Optional[Union[UUID, str]] = Field(
        default=None,
        description="Name/ID of the model associated with the step run.",
    )
    version: Union[int, str, None] = Field(
        default=None,
        description="Version of the step run.",
        union_mode="left_to_right",
    )
    exclude_retried: Optional[bool] = Field(
        default=None,
        description="Whether to exclude retried step runs.",
    )
    cache_expires_at: Optional[Union[datetime, str]] = Field(
        default=None,
        description="Cache expiration time of the step run.",
        union_mode="left_to_right",
    )
    cache_expired: Optional[bool] = Field(
        default=None,
        description="Whether the cache expiration time of the step run has "
        "passed.",
    )
    model_config = ConfigDict(protected_namespaces=())

    def get_custom_filters(
        self, table: Type["AnySchema"]
    ) -> List["ColumnElement[bool]"]:
        """Get custom filters.

        Args:
            table: The query table.

        Returns:
            A list of custom filters.
        """
        custom_filters = super().get_custom_filters(table)

        from sqlmodel import and_, col, or_

        from zenml.zen_stores.schemas import (
            ModelSchema,
            ModelVersionSchema,
            StepRunSchema,
        )

        if self.model:
            model_filter = and_(
                StepRunSchema.model_version_id == ModelVersionSchema.id,
                ModelVersionSchema.model_id == ModelSchema.id,
                self.generate_name_or_id_query_conditions(
                    value=self.model, table=ModelSchema
                ),
            )
            custom_filters.append(model_filter)

        if self.exclude_retried:
            custom_filters.append(
                col(StepRunSchema.status) != ExecutionStatus.RETRIED.value
            )

        if self.cache_expired is True:
            cache_expiration_filter = and_(
                col(StepRunSchema.cache_expires_at).is_not(None),
                col(StepRunSchema.cache_expires_at) < utc_now(),
            )
            custom_filters.append(cache_expiration_filter)
        elif self.cache_expired is False:
            cache_expiration_filter = or_(
                col(StepRunSchema.cache_expires_at) > utc_now(),
                col(StepRunSchema.cache_expires_at).is_(None),
            )
            custom_filters.append(cache_expiration_filter)

        return custom_filters


# ------------------ Heartbeat Model ---------------


class StepHeartbeatResponse(BaseModel, use_enum_values=True):
    """Light-weight model for Step Heartbeat responses."""

    id: UUID
    status: ExecutionStatus
    latest_heartbeat: datetime
    pipeline_run_status: ExecutionStatus | None = None
    heartbeat_enabled: bool



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Abstract base class for entrypoint configurations."""

import argparse
import os
import sys
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List, NoReturn, Optional
from uuid import UUID

from zenml.client import Client
from zenml.code_repositories import BaseCodeRepository
from zenml.enums import StackComponentType
from zenml.exceptions import CustomFlavorImportError
from zenml.logger import get_logger
from zenml.utils import (
    code_repository_utils,
    code_utils,
    source_utils,
    uuid_utils,
)

if TYPE_CHECKING:
    from zenml.artifact_stores import BaseArtifactStore
    from zenml.config import DockerSettings
    from zenml.models import CodeReferenceResponse, PipelineSnapshotResponse

logger = get_logger(__name__)
DEFAULT_ENTRYPOINT_COMMAND = [
    "python",
    "-m",
    "zenml.entrypoints.entrypoint",
]

ENTRYPOINT_CONFIG_SOURCE_OPTION = "entrypoint_config_source"
SNAPSHOT_ID_OPTION = "snapshot_id"


class BaseEntrypointConfiguration(ABC):
    """Abstract base class for entrypoint configurations.

    An entrypoint configuration specifies the arguments that should be passed
    to the entrypoint and what is running inside the entrypoint.

    Attributes:
        entrypoint_args: The parsed arguments passed to the entrypoint.
    """

    def __init__(self, arguments: List[str]):
        """Initializes the entrypoint configuration.

        Args:
            arguments: Command line arguments to configure this object.
        """
        self.entrypoint_args = self._parse_arguments(arguments)
        self._snapshot: Optional["PipelineSnapshotResponse"] = None

    @classmethod
    def get_entrypoint_command(cls) -> List[str]:
        """Returns a command that runs the entrypoint module.

        This entrypoint module is responsible for running the entrypoint
        configuration when called. Defaults to running the
        `zenml.entrypoints.entrypoint` module.

        **Note**: This command won't work on its own but needs to be called with
            the arguments returned by the `get_entrypoint_arguments(...)`
            method of this class.

        Returns:
            A list of strings with the command.
        """
        return DEFAULT_ENTRYPOINT_COMMAND

    @classmethod
    def get_entrypoint_options(cls) -> Dict[str, bool]:
        """Gets all options required for running with this configuration.

        Returns:
            A dictionary of options and whether they are required.
        """
        return {
            # Importable source pointing to the entrypoint configuration class
            # that should be used inside the entrypoint.
            ENTRYPOINT_CONFIG_SOURCE_OPTION: True,
            # ID of the pipeline snapshot to use in this entrypoint
            SNAPSHOT_ID_OPTION: True,
        }

    @classmethod
    def get_entrypoint_arguments(
        cls,
        **kwargs: Any,
    ) -> List[str]:
        """Gets all arguments that the entrypoint command should be called with.

        The argument list should be something that
        `argparse.ArgumentParser.parse_args(...)` can handle (e.g.
        `["--some_option", "some_value"]` or `["--some_option=some_value"]`).
        It needs to provide values for all options returned by the
        `get_entrypoint_options()` method of this class.

        Args:
            **kwargs: Keyword args.

        Returns:
            A list of strings with the arguments.

        Raises:
            ValueError: If no valid snapshot ID is passed.
        """
        arguments = [
            f"--{ENTRYPOINT_CONFIG_SOURCE_OPTION}",
            source_utils.resolve(cls).import_path,
        ]

        if SNAPSHOT_ID_OPTION in cls.get_entrypoint_options():
            snapshot_id = kwargs.get(SNAPSHOT_ID_OPTION)
            if not uuid_utils.is_valid_uuid(snapshot_id):
                raise ValueError(
                    f"Missing or invalid snapshot ID as argument for entrypoint "
                    f"configuration. Please make sure to pass a valid UUID to "
                    f"`{cls.__name__}.{cls.get_entrypoint_arguments.__name__}"
                    f"({SNAPSHOT_ID_OPTION}=<UUID>)`."
                )

            arguments.extend(
                [
                    f"--{SNAPSHOT_ID_OPTION}",
                    str(snapshot_id),
                ]
            )

        return arguments

    @classmethod
    def _parse_arguments(cls, arguments: List[str]) -> Dict[str, Any]:
        """Parses command line arguments.

        This method will create an `argparse.ArgumentParser` and add required
        arguments for all the options specified in the
        `get_entrypoint_options()` method of this class.

        Args:
            arguments: Arguments to parse. The format should be something that
                `argparse.ArgumentParser.parse_args(...)` can handle (e.g.
                `["--some_option", "some_value"]` or
                `["--some_option=some_value"]`).

        Returns:
            Dictionary of the parsed arguments.

        # noqa: DAR402
        Raises:
            ValueError: If the arguments are not valid.
        """

        class _CustomParser(argparse.ArgumentParser):
            """Argument parser subclass that suppresses some argparse logs.

            Also raises an exception instead of the `sys.exit()` call.
            """

            def error(self, message: str) -> NoReturn:
                raise ValueError(
                    f"Failed to parse entrypoint arguments: {message}"
                )

        parser = _CustomParser()

        for option_name, required in cls.get_entrypoint_options().items():
            if option_name == ENTRYPOINT_CONFIG_SOURCE_OPTION:
                # This option is already used by
                # `zenml.entrypoints.entrypoint` to read which config
                # class to use
                continue
            parser.add_argument(f"--{option_name}", required=required)

        result, _ = parser.parse_known_args(arguments)
        return vars(result)

    @property
    def snapshot(self) -> "PipelineSnapshotResponse":
        """The snapshot configured for this entrypoint configuration.

        Returns:
            The snapshot.
        """
        if self._snapshot is None:
            self._snapshot = self._load_snapshot()
        return self._snapshot

    @property
    def docker_settings(self) -> "DockerSettings":
        """The Docker settings configured for this entrypoint configuration.

        Returns:
            The Docker settings.
        """
        return self.snapshot.pipeline_configuration.docker_settings

    @property
    def should_download_code(self) -> bool:
        """Whether code should be downloaded.

        Returns:
            Whether code should be downloaded.
        """
        if (
            self.snapshot.code_reference
            and self.docker_settings.allow_download_from_code_repository
        ):
            return True

        if (
            self.snapshot.code_path
            and self.docker_settings.allow_download_from_artifact_store
        ):
            return True

        return False

    def _load_snapshot(self) -> "PipelineSnapshotResponse":
        """Loads the snapshot.

        Returns:
            The snapshot.
        """
        snapshot_id = UUID(self.entrypoint_args[SNAPSHOT_ID_OPTION])
        return Client().zen_store.get_snapshot(snapshot_id=snapshot_id)

    def prepare_code_environment(self) -> None:
        """Prepares the code environment.

        Raises:
            CustomFlavorImportError: If the artifact store flavor can't be
                imported.
            RuntimeError: If the current environment requires code download
                but the snapshot does not have a reference to any code.
        """
        # Change the working directory to make sure we're in the correct
        # directory where the files in the Docker image should be included.
        # This is necessary as some services overwrite the working directory
        # configured in the Docker image itself.
        os.makedirs("/app", exist_ok=True)
        os.chdir("/app")

        if self.should_download_code:
            if code_path := self.snapshot.code_path:
                # Load the artifact store not from the active stack but separately.
                # This is required in case the stack has custom flavor components
                # (other than the artifact store) for which the flavor
                # implementations will only be available once the download finishes.
                try:
                    artifact_store = self._load_active_artifact_store()
                except CustomFlavorImportError as e:
                    raise CustomFlavorImportError(
                        "Failed to import custom artifact store flavor. The "
                        "artifact store flavor is needed to download your code, "
                        "but it looks like it might be part of the files "
                        "that we're trying to download. If this is the case, you "
                        "should disable downloading code from the artifact store "
                        "using `DockerSettings(allow_download_from_artifact_store=False)` "
                        "or make sure the artifact flavor files are included in "
                        "Docker image by using a custom parent image or installing "
                        "them as part of a pip dependency."
                    ) from e
                code_utils.download_code_from_artifact_store(
                    code_path=code_path, artifact_store=artifact_store
                )
            elif code_reference := self.snapshot.code_reference:
                # TODO: This might fail if the code repository had unpushed changes
                # at the time the pipeline run was started.
                self.download_code_from_code_repository(
                    code_reference=code_reference
                )
            else:
                raise RuntimeError(
                    "Code download required but no code reference or path provided."
                )

            logger.info("Code download finished.")

        # If the working directory is not in the sys.path, we include it to make
        # sure user code gets correctly imported.
        cwd = os.getcwd()
        if cwd not in sys.path:
            sys.path.insert(0, cwd)

    def download_code_from_code_repository(
        self, code_reference: "CodeReferenceResponse"
    ) -> None:
        """Download code from a code repository.

        Args:
            code_reference: The reference to the code.
        """
        logger.info(
            "Downloading code from code repository `%s` (commit `%s`).",
            code_reference.code_repository.name,
            code_reference.commit,
        )

        model = Client().get_code_repository(code_reference.code_repository.id)
        repo = BaseCodeRepository.from_model(model)
        code_repo_root = os.path.abspath("code")
        download_dir = os.path.join(
            code_repo_root, code_reference.subdirectory
        )
        os.makedirs(download_dir, exist_ok=True)
        repo.download_files(
            commit=code_reference.commit,
            directory=download_dir,
            repo_sub_directory=code_reference.subdirectory,
        )
        source_utils.set_custom_source_root(download_dir)
        code_repository_utils.set_custom_local_repository(
            root=code_repo_root, commit=code_reference.commit, repo=repo
        )

        sys.path.insert(0, download_dir)
        os.chdir(download_dir)

    def _load_active_artifact_store(self) -> "BaseArtifactStore":
        """Load the active artifact store.

        Returns:
            The active artifact store.
        """
        from zenml.artifact_stores import BaseArtifactStore

        artifact_store_model = Client().active_stack_model.components[
            StackComponentType.ARTIFACT_STORE
        ][0]
        artifact_store = BaseArtifactStore.from_model(artifact_store_model)
        assert isinstance(artifact_store, BaseArtifactStore)

        return artifact_store

    @abstractmethod
    def run(self) -> None:
        """Runs the entrypoint configuration."""



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Class to launch (run directly or using a step operator) steps."""

import signal
import time
from contextlib import nullcontext
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ContextManager,
    Dict,
    List,
    Optional,
    Tuple,
)

from zenml.client import Client
from zenml.config.step_configurations import Step
from zenml.config.step_run_info import StepRunInfo
from zenml.constants import (
    ENV_ZENML_STEP_OPERATOR,
)
from zenml.enums import ExecutionMode, ExecutionStatus, StepRuntime
from zenml.environment import get_run_environment_dict
from zenml.exceptions import RunInterruptedException, RunStoppedException
from zenml.logger import get_logger
from zenml.models import (
    PipelineRunRequest,
    PipelineRunResponse,
    PipelineSnapshotResponse,
    StepRunResponse,
)
from zenml.models.v2.core.step_run import StepRunInputResponse
from zenml.orchestrators import output_utils, publish_utils, step_run_utils
from zenml.orchestrators import utils as orchestrator_utils
from zenml.orchestrators.step_runner import StepRunner
from zenml.stack import Stack
from zenml.steps import StepHeartBeatTerminationException
from zenml.utils import env_utils, exception_utils, string_utils
from zenml.utils.logging_utils import (
    LoggingContext,
    is_step_logging_enabled,
    setup_logging_context,
)
from zenml.utils.time_utils import utc_now

if TYPE_CHECKING:
    from zenml.step_operators import BaseStepOperator

logger = get_logger(__name__)


def _get_step_operator(
    stack: "Stack", step_operator_name: Optional[str]
) -> "BaseStepOperator":
    """Fetches the step operator from the stack.

    Args:
        stack: Stack on which the step is being run.
        step_operator_name: Name of the step operator to get.

    Returns:
        The step operator to run a step.

    Raises:
        RuntimeError: If no active step operator is found.
    """
    step_operator = stack.step_operator

    # the two following errors should never happen as the stack gets
    # validated before running the pipeline
    if not step_operator:
        raise RuntimeError(
            f"No step operator specified for active stack '{stack.name}'."
        )

    if step_operator_name and step_operator_name != step_operator.name:
        raise RuntimeError(
            f"No step operator named '{step_operator_name}' in active "
            f"stack '{stack.name}'."
        )

    return step_operator


class StepLauncher:
    """A class responsible for launching a step of a ZenML pipeline.

    This class follows these steps to launch and publish a ZenML step:
    1. Publish or reuse a `PipelineRun`
    2. Resolve the input artifacts of the step
    3. Generate a cache key for the step
    4. Check if the step can be cached or not
    5. Publish a new `StepRun`
    6. If the step can't be cached, the step will be executed in one of these
    two ways depending on its configuration:
        - Calling a `step operator` to run the step in a different environment
        - Calling a `step runner` to run the step in the current environment
    7. Update the status of the previously published `StepRun`
    8. Update the status of the `PipelineRun`
    """

    def __init__(
        self,
        snapshot: PipelineSnapshotResponse,
        step: Step,
        orchestrator_run_id: str,
        wait: bool = True,
    ):
        """Initializes the launcher.

        Args:
            snapshot: The pipeline snapshot.
            step: The step to launch.
            orchestrator_run_id: The orchestrator pipeline run id.
            wait: Whether to wait for the step to finish.

        Raises:
            RuntimeError: If the snapshot has no associated stack.
        """
        self._snapshot = snapshot
        self._step = step
        self._orchestrator_run_id = orchestrator_run_id
        self._wait = wait

        if not snapshot.stack:
            raise RuntimeError(
                f"Missing stack for snapshot {snapshot.id}. This is "
                "probably because the stack was manually deleted."
            )

        self._stack = Stack.from_model(snapshot.stack)
        self._invocation_id = step.spec.invocation_id

        # Internal properties and methods
        self._step_run: Optional[StepRunResponse] = None
        self._setup_signal_handlers()

    def _setup_signal_handlers(self) -> None:
        """Set up signal handlers for graceful shutdown, chaining previous handlers."""
        try:
            # Save previous handlers
            self._prev_sigterm_handler = signal.getsignal(signal.SIGTERM)
            self._prev_sigint_handler = signal.getsignal(signal.SIGINT)
        except ValueError as e:
            # This happens when not in the main thread
            logger.debug(f"Cannot set up signal handlers: {e}")
            self._prev_sigterm_handler = None
            self._prev_sigint_handler = None
            return

        def signal_handler(signum: int, frame: Any) -> None:
            """Handle shutdown signals gracefully.

            Args:
                signum: The signal number.
                frame: The frame of the signal handler.

            Raises:
                RunStoppedException: If the pipeline run is stopped by the user.
                RunInterruptedException: If the execution is interrupted for any
                    other reason.
            """
            logger.info(
                f"Received signal shutdown {signum}. Requesting shutdown "
                f"for step '{self._invocation_id}'..."
            )

            try:
                client = Client()

                if self._step_run:
                    pipeline_run = client.get_pipeline_run(
                        self._step_run.pipeline_run_id, hydrate=False
                    )
                else:
                    raise RunInterruptedException(
                        "The execution was interrupted and the step does not "
                        "exist yet."
                    )

                if pipeline_run and pipeline_run.status in [
                    ExecutionStatus.STOPPING,
                    ExecutionStatus.STOPPED,
                ]:
                    if self._step_run:
                        publish_utils.publish_step_run_status_update(
                            step_run_id=self._step_run.id,
                            status=ExecutionStatus.STOPPED,
                            end_time=utc_now(),
                        )
                    raise RunStoppedException("Pipeline run is stopped.")

                step_run = client.get_run_step(
                    self._step_run.id, hydrate=False
                )

                if (
                    pipeline_run.status == ExecutionStatus.FAILED
                    and step_run.status == ExecutionStatus.RUNNING
                    and self._snapshot.pipeline_configuration.execution_mode
                    == ExecutionMode.FAIL_FAST
                ):
                    publish_utils.publish_step_run_status_update(
                        step_run_id=self._step_run.id,
                        status=ExecutionStatus.STOPPED,
                        end_time=utc_now(),
                    )
                    raise RunStoppedException(
                        "Step run was stopped due to a failure in the pipeline "
                        "run and the execution mode 'FAIL_FAST'."
                    )

                elif step_run.status == ExecutionStatus.STOPPING:
                    publish_utils.publish_step_run_status_update(
                        step_run_id=step_run.id,
                        status=ExecutionStatus.STOPPED,
                        end_time=utc_now(),
                    )
                    raise RunStoppedException("Pipeline run is stopped.")
                else:
                    raise RunInterruptedException(
                        "The execution was interrupted."
                    )
            except (RunStoppedException, RunInterruptedException):
                raise
            except Exception as e:
                raise RunInterruptedException(str(e))
            finally:
                # Chain to previous handler if it exists and is not default/ignore
                if signum == signal.SIGTERM and callable(
                    self._prev_sigterm_handler
                ):
                    self._prev_sigterm_handler(signum, frame)
                elif signum == signal.SIGINT and callable(
                    self._prev_sigint_handler
                ):
                    self._prev_sigint_handler(signum, frame)

        # Register handlers for common termination signals
        try:
            signal.signal(signal.SIGTERM, signal_handler)
            signal.signal(signal.SIGINT, signal_handler)
        except ValueError as e:
            # This happens when not in the main thread
            logger.debug(f"Cannot register signal handlers: {e}")
            # Continue without signal handling - the step will still run

    def launch(self) -> StepRunResponse:
        """Launches the step.

        Raises:
            RunStoppedException: If the pipeline run is stopped by the user.
            BaseException: If the step preparation or execution fails.
            RuntimeError: If an unexpected step run status is encountered.

        Returns:
            The step run response.
        """
        publish_utils.step_exception_info.set(None)

        logs_context: ContextManager[Any] = nullcontext()

        if is_step_logging_enabled(
            step_configuration=self._step.config,
            pipeline_configuration=self._snapshot.pipeline_configuration,
        ):
            logs_context = setup_logging_context(source="prepare_step")

        with logs_context:
            pipeline_run, run_was_created = self._create_or_reuse_run()

            if isinstance(logs_context, LoggingContext):
                logs_context.update(pipeline_run=pipeline_run)

            if run_was_created:
                pipeline_run_metadata = self._stack.get_pipeline_run_metadata(
                    run_id=pipeline_run.id
                )
                publish_utils.publish_pipeline_run_metadata(
                    pipeline_run_id=pipeline_run.id,
                    pipeline_run_metadata=pipeline_run_metadata,
                )
                if model_version := pipeline_run.model_version:
                    step_run_utils.log_model_version_dashboard_url(
                        model_version=model_version
                    )

            request_factory = step_run_utils.StepRunRequestFactory(
                snapshot=self._snapshot,
                pipeline_run=pipeline_run,
                stack=self._stack,
            )
            dynamic_config = self._step if self._snapshot.is_dynamic else None
            step_run_request = request_factory.create_request(
                invocation_id=self._invocation_id,
                dynamic_config=dynamic_config,
            )
            if isinstance(logs_context, LoggingContext):
                step_run_request.logs = logs_context.log_model.id

            try:
                request_factory.populate_request(request=step_run_request)
            except BaseException as e:
                logger.exception(
                    f"Failed preparing step `{self._invocation_id}`."
                )
                step_run_request.status = ExecutionStatus.FAILED
                step_run_request.end_time = utc_now()
                step_run_request.exception_info = (
                    exception_utils.collect_exception_information(e)
                )
                raise
            finally:
                step_run = Client().zen_store.create_run_step(step_run_request)

                if isinstance(logs_context, LoggingContext):
                    logs_context.update(step_run=step_run)

                self._step_run = step_run
                if model_version := step_run.model_version:
                    step_run_utils.log_model_version_dashboard_url(
                        model_version=model_version
                    )

                if not step_run.status.is_finished:
                    logger.info(f"Step `{self._invocation_id}` has started.")

                    start_time = time.time()
                    try:
                        self._run_step(
                            pipeline_run=pipeline_run,
                            step_run=step_run,
                            force_write_logs=lambda: None,
                        )
                    except RunStoppedException as e:
                        raise e
                    except BaseException as e:  # noqa: E722
                        step_run = Client().get_run_step(
                            step_run_id=step_run.id
                        )

                        if (
                            isinstance(e, StepHeartBeatTerminationException)
                            or step_run.status == ExecutionStatus.STOPPING
                        ):
                            # Handle as a non-failure as exception is a propagation of graceful termination.
                            publish_utils.publish_stopped_step_run(step_run.id)

                        else:
                            logger.error(
                                "Failed to run step `%s`: %s",
                                self._invocation_id,
                                e,
                            )
                            if step_run.status == ExecutionStatus.RUNNING:
                                # Only update the status if the step runner
                                # somehow failed to do so.
                                publish_utils.publish_failed_step_run(
                                    step_run.id
                                )
                        raise

                    if self._wait:
                        duration = time.time() - start_time
                        logger.info(
                            f"Step `{self._invocation_id}` has finished in "
                            f"`{string_utils.get_human_readable_time(duration)}`."
                        )
                    else:
                        logger.info("Step `%s` launched.", self._invocation_id)
                elif step_run.status == ExecutionStatus.CACHED:
                    logger.info(
                        f"Using cached version of step `{self._invocation_id}`."
                    )
                    if (
                        model_version := step_run.model_version
                        or pipeline_run.model_version
                    ):
                        step_run_utils.link_output_artifacts_to_model_version(
                            artifacts=step_run.outputs,
                            model_version=model_version,
                        )
                elif step_run.status == ExecutionStatus.SKIPPED:
                    logger.info("Skipping step `%s`.", self._invocation_id)
                    if (
                        model_version := step_run.model_version
                        or pipeline_run.model_version
                    ):
                        step_run_utils.link_output_artifacts_to_model_version(
                            artifacts=step_run.outputs,
                            model_version=model_version,
                        )
                elif step_run.status == ExecutionStatus.FAILED:
                    # No need to link anything for a failed step run.
                    pass
                else:
                    raise RuntimeError(
                        f"Unexpected step run status `{step_run.status}` for "
                        f"step `{self._invocation_id}`."
                    )

        return step_run

    def _create_or_reuse_run(self) -> Tuple[PipelineRunResponse, bool]:
        """Creates a pipeline run or reuses an existing one.

        Returns:
            The created or existing pipeline run,
            and a boolean indicating whether the run was created or reused.
        """
        start_time = utc_now()
        run_name = string_utils.format_name_template(
            name_template=self._snapshot.run_name_template,
            substitutions=self._snapshot.pipeline_configuration.finalize_substitutions(
                start_time=start_time,
            ),
        )

        logger.debug("Creating pipeline run %s", run_name)

        client = Client()
        pipeline_run = PipelineRunRequest(
            name=run_name,
            orchestrator_run_id=self._orchestrator_run_id,
            project=client.active_project.id,
            snapshot=self._snapshot.id,
            status=ExecutionStatus.RUNNING,
            orchestrator_environment=get_run_environment_dict(),
            start_time=start_time,
            tags=self._snapshot.pipeline_configuration.tags,
        )
        return client.zen_store.get_or_create_run(pipeline_run)

    def _run_step(
        self,
        pipeline_run: PipelineRunResponse,
        step_run: StepRunResponse,
        force_write_logs: Callable[..., Any],
    ) -> None:
        """Runs the current step.

        Args:
            pipeline_run: The model of the current pipeline run.
            step_run: The model of the current step run.
            force_write_logs: The context for the step logs.

        """
        from zenml.deployers.server import runtime

        step_run_info = StepRunInfo(
            step_run=step_run,
            config=self._step.config,
            spec=self._step.spec,
            pipeline=self._snapshot.pipeline_configuration,
            run_name=pipeline_run.name,
            pipeline_step_name=self._invocation_id,
            run_id=pipeline_run.id,
            step_run_id=step_run.id,
            force_write_logs=force_write_logs,
            snapshot=self._snapshot,
        )

        output_artifact_uris = output_utils.prepare_output_artifact_uris(
            step_run=step_run,
            stack=self._stack,
            step=self._step,
            skip_artifact_materialization=runtime.should_skip_artifact_materialization(),
        )

        try:
            if self._step.config.step_operator:
                step_operator_name = None
                if isinstance(self._step.config.step_operator, str):
                    step_operator_name = self._step.config.step_operator

                self._run_step_with_step_operator(
                    step_operator_name=step_operator_name,
                    step_run_info=step_run_info,
                )
            elif not self._snapshot.is_dynamic:
                self._run_step_in_current_thread(
                    pipeline_run=pipeline_run,
                    step_run=step_run,
                    step_run_info=step_run_info,
                    input_artifacts=step_run.regular_inputs,
                    output_artifact_uris=output_artifact_uris,
                )
            else:
                from zenml.execution.pipeline.dynamic.runner import (
                    get_step_runtime,
                )

                step_runtime = get_step_runtime(
                    step_config=self._step.config,
                    pipeline_docker_settings=self._snapshot.pipeline_configuration.docker_settings,
                    orchestrator=self._stack.orchestrator,
                )

                if step_runtime == StepRuntime.INLINE:
                    if self._step.config.runtime == StepRuntime.ISOLATED:
                        # The step was configured to run in an isolated runtime,
                        # but the orchestrator doesn't support it.
                        logger.warning(
                            "The %s does not support running steps "
                            "in isolated runtimes. Running step `%s` in inline "
                            "runtime instead.",
                            self._stack.orchestrator.__class__.__name__,
                            self._invocation_id,
                        )

                    self._run_step_in_current_thread(
                        pipeline_run=pipeline_run,
                        step_run=step_run,
                        step_run_info=step_run_info,
                        input_artifacts=step_run.regular_inputs,
                        output_artifact_uris=output_artifact_uris,
                    )
                else:
                    self._run_step_with_dynamic_orchestrator(
                        step_run_info=step_run_info
                    )
        except:  # noqa: E722
            output_utils.remove_artifact_dirs(
                artifact_uris=list(output_artifact_uris.values())
            )
            raise

    def _run_step_with_step_operator(
        self,
        step_operator_name: Optional[str],
        step_run_info: StepRunInfo,
    ) -> None:
        """Runs the current step with a step operator.

        Args:
            step_operator_name: The name of the step operator to use.
            step_run_info: Additional information needed to run the step.

        # noqa: DAR401
        Raises:
            RuntimeError: If trying to use a step operator that does not support
                running asynchronously in a dynamic pipeline.
            NotImplementedError: If the step operator does not implement the
                `submit(...)` or `launch(...)` methods.
            BaseException: If the step run failed.
        """
        step_operator = _get_step_operator(
            stack=self._stack,
            step_operator_name=step_operator_name,
        )
        entrypoint_cfg_class = step_operator.entrypoint_config_class
        entrypoint_command = (
            entrypoint_cfg_class.get_entrypoint_command()
            + entrypoint_cfg_class.get_entrypoint_arguments(
                step_name=self._invocation_id,
                snapshot_id=self._snapshot.id,
                step_run_id=str(step_run_info.step_run_id),
            )
        )
        environment, secrets = orchestrator_utils.get_config_environment_vars(
            pipeline_run_id=step_run_info.run_id,
        )
        # TODO: for now, we don't support separate secrets from environment
        # in the step operator environment
        environment.update(secrets)

        environment.update(
            env_utils.get_runtime_environment(
                config=step_run_info.config,
                stack=self._stack,
            )
        )
        environment[ENV_ZENML_STEP_OPERATOR] = "True"
        logger.info(
            "Using step operator `%s` to run step `%s`.",
            step_operator.name,
            self._invocation_id,
        )

        try:
            step_operator.submit(
                info=step_run_info,
                entrypoint_command=entrypoint_command,
                environment=environment,
            )
        except NotImplementedError:
            if not self._wait:
                # We're running in a dynamic pipeline and for the monitoring to
                # work correctly, we only allow running with step operators that
                # support running asynchronously.
                raise RuntimeError(
                    f"The step operator `{step_operator.name}` does not "
                    "support running asynchronously and therefore cannot be "
                    "used in dynamic pipelines. To solve this, please "
                    "implement the `submit(...)` and `get_status(...)` methods "
                    "in the step operator class."
                )

            if hasattr(step_operator, "launch") and callable(
                step_operator.launch
            ):
                # We fallback to the legacy launch method if it is implemented.
                # The launch method is synchronous and will block until the step
                # run is finished, so there is no need to wait for it.
                logger.debug(
                    "Falling back to the legacy launch method for "
                    "step operator `%s`.",
                    step_operator.name,
                )
                step_operator.launch(
                    info=step_run_info,
                    entrypoint_command=entrypoint_command,
                    environment=environment,
                )
            else:
                raise NotImplementedError(
                    f"The step operator `{step_operator.name}` does not "
                    "implement the `submit(...)` or `launch(...)` methods."
                )
        else:
            # We submitted the step run asynchronously, now we potentially need
            # to wait for it to finish.
            if self._wait:
                status = step_operator.wait(
                    step_run=step_run_info.step_run,
                )
                if not status.is_successful:
                    step_run = Client().get_run_step(step_run_info.step_run_id)
                    raise exception_utils.reconstruct_exception(
                        exception_info=step_run.exception_info,
                        fallback_message=(
                            f"Step `{step_run_info.pipeline_step_name}` failed "
                            f"with status `{status}`."
                        ),
                    )

    def _run_step_with_dynamic_orchestrator(
        self,
        step_run_info: StepRunInfo,
    ) -> None:
        """Runs the current step with a dynamic orchestrator.

        Args:
            step_run_info: Additional information needed to run the step.

        # noqa: DAR401
        Raises:
            BaseException: If the step run failed.
        """
        # If we don't pass the run ID here, does it reuse the existing token?
        environment, secrets = orchestrator_utils.get_config_environment_vars(
            pipeline_run_id=step_run_info.run_id,
        )
        environment.update(secrets)

        environment.update(
            env_utils.get_runtime_environment(
                config=step_run_info.config,
                stack=self._stack,
            )
        )
        self._stack.orchestrator.submit_isolated_step(
            step_run_info=step_run_info,
            environment=environment,
        )
        if self._wait:
            status = self._stack.orchestrator.wait_for_isolated_step(
                step_run_info.step_run
            )
            if not status.is_successful:
                step_run = Client().get_run_step(step_run_info.step_run_id)
                raise exception_utils.reconstruct_exception(
                    exception_info=step_run.exception_info,
                    fallback_message=(
                        f"Step `{step_run_info.pipeline_step_name}` failed "
                        f"with status `{status}`."
                    ),
                )

    def _run_step_in_current_thread(
        self,
        pipeline_run: PipelineRunResponse,
        step_run: StepRunResponse,
        step_run_info: StepRunInfo,
        input_artifacts: Dict[str, List["StepRunInputResponse"]],
        output_artifact_uris: Dict[str, str],
    ) -> None:
        """Runs the current step without a step operator.

        Args:
            pipeline_run: The model of the current pipeline run.
            step_run: The model of the current step run.
            step_run_info: Additional information needed to run the step.
            input_artifacts: The input artifact versions of the current step.
            output_artifact_uris: The output artifact URIs of the current step.
        """
        runner = StepRunner(step=self._step, stack=self._stack)
        runner.run(
            pipeline_run=pipeline_run,
            step_run=step_run,
            input_artifacts=input_artifacts,
            output_artifact_uris=output_artifact_uris,
            step_run_info=step_run_info,
        )



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Models representing the link between model versions and pipeline runs."""

from typing import TYPE_CHECKING, List, Optional, Type, TypeVar, Union
from uuid import UUID

from pydantic import ConfigDict, Field

from zenml.enums import GenericFilterOps
from zenml.models.v2.base.base import (
    BaseDatedResponseBody,
    BaseIdentifiedResponse,
    BaseRequest,
    BaseResponseMetadata,
    BaseResponseResources,
)
from zenml.models.v2.base.filter import BaseFilter, StrFilter
from zenml.models.v2.core.pipeline_run import PipelineRunResponse

if TYPE_CHECKING:
    from sqlalchemy.sql.elements import ColumnElement

    from zenml.zen_stores.schemas import BaseSchema

    AnySchema = TypeVar("AnySchema", bound=BaseSchema)


# ------------------ Request Model ------------------


class ModelVersionPipelineRunRequest(BaseRequest):
    """Request model for links between model versions and pipeline runs."""

    model_version: UUID
    pipeline_run: UUID

    # TODO: In Pydantic v2, the `model_` is a protected namespaces for all
    #  fields defined under base models. If not handled, this raises a warning.
    #  It is possible to suppress this warning message with the following
    #  configuration, however the ultimate solution is to rename these fields.
    #  Even though they do not cause any problems right now, if we are not
    #  careful we might overwrite some fields protected by pydantic.
    model_config = ConfigDict(protected_namespaces=())


# ------------------ Update Model ------------------

# There is no update model for links between model version and pipeline runs.

# ------------------ Response Model ------------------


class ModelVersionPipelineRunResponseBody(BaseDatedResponseBody):
    """Response body for links between model versions and pipeline runs."""

    model_version: UUID
    pipeline_run: PipelineRunResponse

    # TODO: In Pydantic v2, the `model_` is a protected namespaces for all
    #  fields defined under base models. If not handled, this raises a warning.
    #  It is possible to suppress this warning message with the following
    #  configuration, however the ultimate solution is to rename these fields.
    #  Even though they do not cause any problems right now, if we are not
    #  careful we might overwrite some fields protected by pydantic.
    model_config = ConfigDict(protected_namespaces=())


class ModelVersionPipelineRunResponseResources(BaseResponseResources):
    """Class for all resource models associated with the model version pipeline run entity."""


class ModelVersionPipelineRunResponse(
    BaseIdentifiedResponse[
        ModelVersionPipelineRunResponseBody,
        BaseResponseMetadata,
        ModelVersionPipelineRunResponseResources,
    ]
):
    """Response model for links between model versions and pipeline runs."""

    @property
    def model_version(self) -> UUID:
        """The `model_version` property.

        Returns:
            the value of the property.
        """
        return self.get_body().model_version

    @property
    def pipeline_run(self) -> "PipelineRunResponse":
        """The `pipeline_run` property.

        Returns:
            the value of the property.
        """
        return self.get_body().pipeline_run


# ------------------ Filter Model ------------------


class ModelVersionPipelineRunFilter(BaseFilter):
    """Model version pipeline run links filter model."""

    FILTER_EXCLUDE_FIELDS = [
        *BaseFilter.FILTER_EXCLUDE_FIELDS,
        "pipeline_run_name",
        "user",
    ]
    CLI_EXCLUDE_FIELDS = [
        *BaseFilter.CLI_EXCLUDE_FIELDS,
        "model_version_id",
        "updated",
        "id",
    ]

    model_version_id: Optional[Union[UUID, str]] = Field(
        default=None,
        description="Filter by model version ID",
        union_mode="left_to_right",
    )
    pipeline_run_id: Optional[Union[UUID, str]] = Field(
        default=None,
        description="Filter by pipeline run ID",
        union_mode="left_to_right",
    )
    pipeline_run_name: Optional[str] = Field(
        default=None,
        description="Name of the pipeline run",
    )
    user: Optional[Union[UUID, str]] = Field(
        default=None,
        description="Name/ID of the user that created the pipeline run.",
    )

    # TODO: In Pydantic v2, the `model_` is a protected namespaces for all
    #  fields defined under base models. If not handled, this raises a warning.
    #  It is possible to suppress this warning message with the following
    #  configuration, however the ultimate solution is to rename these fields.
    #  Even though they do not cause any problems right now, if we are not
    #  careful we might overwrite some fields protected by pydantic.
    model_config = ConfigDict(protected_namespaces=())

    def get_custom_filters(
        self, table: Type["AnySchema"]
    ) -> List["ColumnElement[bool]"]:
        """Get custom filters.

        Args:
            table: The query table.

        Returns:
            A list of custom filters.
        """
        custom_filters = super().get_custom_filters(table)

        from sqlmodel import and_

        from zenml.zen_stores.schemas import (
            ModelVersionPipelineRunSchema,
            PipelineRunSchema,
            UserSchema,
        )

        if self.pipeline_run_name:
            value, filter_operator = self._resolve_operator(
                self.pipeline_run_name
            )
            filter_ = StrFilter(
                operation=GenericFilterOps(filter_operator),
                column="name",
                value=value,
            )
            pipeline_run_name_filter = and_(
                ModelVersionPipelineRunSchema.pipeline_run_id
                == PipelineRunSchema.id,
                filter_.generate_query_conditions(PipelineRunSchema),
            )
            custom_filters.append(pipeline_run_name_filter)

        if self.user:
            user_filter = and_(
                ModelVersionPipelineRunSchema.pipeline_run_id
                == PipelineRunSchema.id,
                PipelineRunSchema.user_id == UserSchema.id,
                self.generate_name_or_id_query_conditions(
                    value=self.user,
                    table=UserSchema,
                    additional_columns=["full_name"],
                ),
            )
            custom_filters.append(user_filter)

        return custom_filters



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:

#       https://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Functionality to install or uninstall ZenML integrations via the CLI."""

import os
import subprocess
import sys
from typing import Optional, Tuple

import click
from rich.progress import track

from zenml.cli.cli import TagGroup, cli
from zenml.cli.utils import (
    confirmation,
    declare,
    error,
    exception,
    format_integration_list,
    install_packages,
    print_table,
    title,
    uninstall_package,
    warning,
)
from zenml.console import console
from zenml.enums import CliCategories
from zenml.logger import get_logger

_WHYLOGS_INTEGRATION_WARNING = (
    "WhyLabs was acquired by Apple and the hosted WhyLabs platform is being "
    "discontinued. The whylogs library remains open source and the WhyLabs "
    "platform is now available as OSS at "
    "https://github.com/whylabs/whylabs-oss, but hosted functionality may stop "
    "working. Plan accordingly before continuing with the whylogs integration."
)

logger = get_logger(__name__)


@cli.group(
    cls=TagGroup,
    tag=CliCategories.INTEGRATIONS,
)
def integration() -> None:
    """Interact with external integrations."""


@integration.command(name="list", help="List the available integrations.")
def list_integrations() -> None:
    """List all available integrations with their installation status."""
    from zenml.integrations.registry import integration_registry

    formatted_table = format_integration_list(
        sorted(list(integration_registry.integrations.items()))
    )
    print_table(formatted_table)
    warning(
        "\n" + "To install the dependencies of a specific integration, type: "
    )
    warning("zenml integration install INTEGRATION_NAME")


@integration.command(
    name="requirements", help="List all requirements for an integration."
)
@click.argument("integration_name", required=False, default=None)
def get_requirements(integration_name: Optional[str] = None) -> None:
    """List all requirements for the chosen integration.

    Args:
        integration_name: The name of the integration to list the requirements
            for.
    """
    from zenml.integrations.registry import integration_registry

    try:
        requirements = integration_registry.select_integration_requirements(
            integration_name
        )
    except KeyError as e:
        exception(e)
    else:
        if requirements:
            title(
                f"Requirements for {integration_name or 'all integrations'}:\n"
            )
            declare(f"{requirements}")
            warning(
                "\n" + "To install the dependencies of a "
                "specific integration, type: "
            )
            warning("zenml integration install INTEGRATION_NAME")


@integration.command(
    name="export-requirements", help="Export the integration requirements."
)
@click.argument("integrations", nargs=-1, required=False)
@click.option(
    "--ignore-integration",
    "-i",
    multiple=True,
    help="List of integrations to ignore explicitly.",
)
@click.option(
    "--output-file",
    "-o",
    "output_file",
    type=str,
    required=False,
    help="File to which to export the integration requirements. If not "
    "provided, the requirements will be printed to stdout instead.",
)
@click.option(
    "--overwrite",
    "-ov",
    "overwrite",
    type=bool,
    required=False,
    is_flag=True,
    help="Overwrite the output file if it already exists. This option is "
    "only valid if the output file is provided.",
)
@click.option(
    "--installed-only",
    "installed_only",
    is_flag=True,
    default=False,
    help="Only export requirements for integrations installed in your current "
    "environment. This can not be specified when also providing explicit "
    "integrations.",
)
@click.option(
    "--poetry",
    "poetry",
    is_flag=True,
    default=False,
    help="Add the exported requirements to your current Poetry project.",
)
def export_requirements(
    integrations: Tuple[str],
    ignore_integration: Tuple[str],
    output_file: Optional[str] = None,
    overwrite: bool = False,
    installed_only: bool = False,
    poetry: bool = False,
) -> None:
    """Exports integration requirements so they can be installed using pip.

    Args:
        integrations: The name of the integration to install the requirements
            for.
        ignore_integration: List of integrations to ignore explicitly.
        output_file: Optional path to the requirements output file.
        overwrite: Overwrite the output file if it already exists. This option
            is only valid if the output file is provided.
        installed_only: Only export requirements for integrations installed in
            your current environment. This can not be specified when also
            providing explicit integrations.
        poetry: Add the exported requirements to your current Poetry project.
    """
    from zenml.integrations.registry import integration_registry

    if installed_only and integrations:
        error(
            "You can either provide specific integrations or export only "
            "requirements for integrations installed in your local "
            "environment, not both."
        )

    if poetry and output_file:
        error(
            "You can either specify an output file or add the requirements to "
            "the Poetry project, not both."
        )

    all_integrations = set(integration_registry.integrations.keys())

    if integrations:
        integrations_to_export = set(integrations)
    elif installed_only:
        integrations_to_export = set(
            integration_registry.get_installed_integrations()
        )
    else:
        integrations_to_export = all_integrations

    for i in ignore_integration:
        try:
            integrations_to_export.remove(i)
        except KeyError:
            if i not in all_integrations:
                error(
                    f"Integration {i} does not exist. Available integrations: "
                    f"{all_integrations}"
                )

    if "whylogs" in integrations_to_export:
        warning(_WHYLOGS_INTEGRATION_WARNING)

    requirements = []
    for integration_name in integrations_to_export:
        try:
            requirements += (
                integration_registry.select_integration_requirements(
                    integration_name
                )
            )
        except KeyError:
            error(f"Unable to find integration '{integration_name}'.")

    if output_file:
        try:
            with open(output_file, "x") as f:
                f.write("\n".join(requirements))
        except FileExistsError:
            if overwrite or confirmation(
                "A file already exists at the specified path. "
                "Would you like to overwrite it?"
            ):
                with open(output_file, "w") as f:
                    f.write("\n".join(requirements))
        declare(f"Requirements exported to {output_file}.")
    if poetry:
        res = os.popen("poetry env list").read()
        envs = [
            env
            for env in res.split("\n")
            if env.lower().find("(activated)") > 0
        ]
        if len(envs) == 0:
            error(
                "No activated Poetry environment found. Please activate one "
                "and try again."
            )
        else:
            # Use subprocess.run with shell=False to avoid command injection
            args = ["poetry", "add"] + requirements
            subprocess.run(args, check=True)
            declare(
                f"Requirements added to `{envs[0]}` environment in Poetry."
            )
    else:
        click.echo(" ".join(requirements), nl=False)


@integration.command(
    help="Install the required packages for the integration of choice."
)
@click.argument("integrations", nargs=-1, required=False)
@click.option(
    "--ignore-integration",
    "-i",
    multiple=True,
    help="Integrations to ignore explicitly (passed in separately).",
)
@click.option(
    "--yes",
    "-y",
    "force",
    is_flag=True,
    help="Force the installation of the required packages. This will skip the "
    "confirmation step and reinstall existing packages as well",
)
@click.option(
    "--uv",
    "uv",
    is_flag=True,
    help="Experimental: Use uv for package installation.",
    default=False,
)
def install(
    integrations: Tuple[str],
    ignore_integration: Tuple[str],
    force: bool = False,
    uv: bool = False,
) -> None:
    """Installs the required packages for a given integration.

    If no integration is specified all required packages for all integrations
    are installed using pip or uv.

    Args:
        integrations: The name of the integration to install the requirements
            for.
        ignore_integration: Integrations to ignore explicitly (passed in
            separately).
        force: Force the installation of the required packages.
        uv: Use uv for package installation (experimental).
    """
    from zenml.cli.utils import is_pip_installed, is_uv_installed
    from zenml.integrations.registry import integration_registry

    if uv and not is_uv_installed():
        error(
            "UV is not installed but the uv flag was passed in. Please install "
            "uv or remove the uv flag."
        )

    if not uv and not is_pip_installed():
        error(
            "Pip is not installed. Please install pip or use the uv flag "
            "(--uv) for package installation."
        )

    if not integrations:
        # no integrations specified, use all registered integrations
        integration_set = set(integration_registry.integrations.keys())

        for i in ignore_integration:
            try:
                integration_set.remove(i)
            except KeyError:
                error(
                    f"Integration {i} does not exist. Available integrations: "
                    f"{list(integration_registry.integrations.keys())}"
                )
    else:
        integration_set = set(integrations)

    if "whylogs" in integration_set:
        warning(_WHYLOGS_INTEGRATION_WARNING)

    if sys.version_info.minor == 12 and "tensorflow" in integration_set:
        warning(
            "The TensorFlow integration is not yet compatible with Python "
            "3.12, thus its installation is skipped. Consider using a "
            "different version of Python and stay in touch for further updates."
        )
        integration_set.remove("tensorflow")

    if sys.version_info.minor == 12 and "deepchecks" in integration_set:
        warning(
            "The Deepchecks integration is not yet compatible with Python "
            "3.12, thus its installation is skipped. Consider using a "
            "different version of Python and stay in touch for further updates."
        )
        integration_set.remove("deepchecks")

    requirements = []
    integrations_to_install = []
    for integration_name in integration_set:
        try:
            if force or not integration_registry.is_installed(
                integration_name
            ):
                requirements += (
                    integration_registry.select_integration_requirements(
                        integration_name
                    )
                )
                integrations_to_install.append(integration_name)
            else:
                declare(
                    f"All required packages for integration "
                    f"'{integration_name}' are already installed."
                )
        except KeyError:
            error(f"Unable to find integration '{integration_name}'.")

    if requirements and (
        force
        or confirmation(
            "Are you sure you want to install the following "
            "packages to the current environment?\n"
            f"{requirements}"
        )
    ):
        with console.status("Installing integrations..."):
            install_packages(requirements, use_uv=uv)


@integration.command(
    help="Uninstall the required packages for the integration of choice."
)
@click.argument("integrations", nargs=-1, required=False)
@click.option(
    "--yes",
    "-y",
    "force",
    is_flag=True,
    help="Force the uninstallation of the required packages. This will skip "
    "the confirmation step",
)
@click.option(
    "--uv",
    "uv",
    is_flag=True,
    help="Experimental: Use uv for package uninstallation.",
    default=False,
)
def uninstall(
    integrations: Tuple[str], force: bool = False, uv: bool = False
) -> None:
    """Uninstalls the required packages for a given integration.

    If no integration is specified all required packages for all integrations
    are uninstalled using pip or uv.

    Args:
        integrations: The name of the integration to uninstall the requirements
            for.
        force: Force the uninstallation of the required packages.
        uv: Use uv for package uninstallation (experimental).
    """
    from zenml.cli.utils import is_pip_installed, is_uv_installed
    from zenml.integrations.registry import integration_registry

    if uv and not is_uv_installed():
        error("Package `uv` is not installed. Please install it and retry.")

    if not uv and not is_pip_installed():
        error(
            "Pip is not installed. Please install pip or use the uv flag "
            "(--uv) for package installation."
        )

    if not integrations:
        # no integrations specified, use all registered integrations
        integrations = tuple(integration_registry.integrations.keys())

    requirements = []
    for integration_name in integrations:
        try:
            if integration_registry.is_installed(integration_name):
                requirements += (
                    integration_registry.select_uninstall_requirements(
                        integration_name
                    )
                )
            else:
                warning(
                    f"Requirements for integration '{integration_name}' "
                    f"already not installed."
                )
        except KeyError:
            warning(f"Unable to find integration '{integration_name}'.")

    if requirements and (
        force
        or confirmation(
            "Are you sure you want to uninstall the following "
            "packages from the current environment?\n"
            f"{requirements}"
        )
    ):
        for n in track(
            range(len(requirements)),
            description="Uninstalling integrations...",
        ):
            uninstall_package(requirements[n], use_uv=uv)


@integration.command(
    help="Upgrade the required packages for the integration of choice."
)
@click.argument("integrations", nargs=-1, required=False)
@click.option(
    "--yes",
    "-y",
    "force",
    is_flag=True,
    help="Force the upgrade of the required packages. This will skip the "
    "confirmation step and re-upgrade existing packages as well",
)
@click.option(
    "--uv",
    "uv",
    is_flag=True,
    help="Experimental: Use uv for package upgrade.",
    default=False,
)
def upgrade(
    integrations: Tuple[str],
    force: bool = False,
    uv: bool = False,
) -> None:
    """Upgrade the required packages for a given integration.

    If no integration is specified all required packages for all integrations
    are installed using pip or uv.

    Args:
        integrations: The name of the integration to install the requirements
            for.
        force: Force the installation of the required packages.
        uv: Use uv for package installation (experimental).
    """
    from zenml.cli.utils import is_pip_installed, is_uv_installed
    from zenml.integrations.registry import integration_registry

    if uv and not is_uv_installed():
        error("Package `uv` is not installed. Please install it and retry.")

    if not uv and not is_pip_installed():
        error(
            "Pip is not installed. Please install pip or use the uv flag "
            "(--uv) for package installation."
        )

    if not integrations:
        # no integrations specified, use all registered integrations
        integrations = set(integration_registry.integrations.keys())

    requirements = []
    integrations_to_install = []
    for integration_name in integrations:
        try:
            if integration_registry.is_installed(integration_name):
                requirements += (
                    integration_registry.select_integration_requirements(
                        integration_name
                    )
                )
                integrations_to_install.append(integration_name)
            else:
                declare(
                    f"None of the required packages for integration "
                    f"'{integration_name}' are installed."
                )
        except KeyError:
            warning(f"Unable to find integration '{integration_name}'.")

    if requirements and (
        force
        or confirmation(
            f"Are you sure you want to upgrade the following "
            "packages to the current environment?\n"
            f"{requirements}"
        )
    ):
        with console.status("Upgrading integrations..."):
            install_packages(requirements, upgrade=True, use_uv=uv)



"""add default deployer [279d55228d28].

Revision ID: 279d55228d28
Revises: 124b57b8c7b1
Create Date: 2025-10-23 18:12:26.232710

"""

import base64
from uuid import uuid4

import sqlalchemy as sa
from alembic import op

from zenml.utils.time_utils import utc_now

# revision identifiers, used by Alembic.
revision = "279d55228d28"
down_revision = "124b57b8c7b1"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database data to add a default deployer component."""
    connection = op.get_bind()

    meta = sa.MetaData()
    meta.reflect(
        bind=connection,
        only=("stack", "stack_component", "stack_composition"),
    )

    stack_table = sa.Table("stack", meta)
    component_table = sa.Table("stack_component", meta)
    composition_table = sa.Table("stack_composition", meta)

    default_stack = connection.execute(
        sa.select(stack_table.c.id)
        .where(stack_table.c.name == "default")
        .limit(1)
    ).fetchone()

    if not default_stack:
        return

    default_stack_id = default_stack[0]
    if not default_stack_id:
        return

    now = utc_now()
    empty_config = base64.b64encode("{}".encode("utf-8"))

    # If an existing default deployer exists, rename it to avoid conflicts
    existing_default = connection.execute(
        sa.select(component_table.c.id)
        .where(component_table.c.name == "default")
        .where(component_table.c.type == "deployer")
        .limit(1)
    ).fetchone()

    if existing_default:
        existing_id = existing_default[0]
        suffix = str(existing_id)[:8]
        connection.execute(
            sa.update(component_table)
            .where(component_table.c.id == existing_id)
            .values(name=f"default_{suffix}", updated=now)
        )

    component_id = uuid4().hex
    connection.execute(
        sa.insert(component_table).values(
            id=component_id,
            name="default",
            type="deployer",
            flavor="local",
            configuration=empty_config,
            created=now,
            updated=now,
        )
    )

    connection.execute(
        sa.insert(composition_table).values(
            stack_id=default_stack_id, component_id=component_id
        )
    )


def downgrade() -> None:
    """No downgrade supported for this data migration."""
    pass



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""OpenTelemetry log store flavor."""

from typing import Dict, Optional, Type

from pydantic import Field

from zenml import __version__
from zenml.constants import (
    LOGS_OTEL_EXPORT_TIMEOUT_MILLIS,
    LOGS_OTEL_MAX_EXPORT_BATCH_SIZE,
    LOGS_OTEL_MAX_QUEUE_SIZE,
    LOGS_OTEL_SCHEDULE_DELAY_MILLIS,
)
from zenml.enums import StackComponentType
from zenml.log_stores import BaseLogStoreConfig
from zenml.log_stores.base_log_store import BaseLogStore
from zenml.stack.flavor import Flavor
from zenml.utils.enum_utils import StrEnum


class Compression(StrEnum):
    """Compression types."""

    NoCompression = "none"
    Deflate = "deflate"
    Gzip = "gzip"


class OtelLogStoreConfig(BaseLogStoreConfig):
    """Configuration for OpenTelemetry log store.

    Attributes:
        service_name: Name of the service (defaults to "zenml").
        service_version: Version of the service (defaults to the ZenML version).
        max_queue_size: Maximum queue size for batch processor.
        schedule_delay_millis: Delay between batch exports in milliseconds.
        max_export_batch_size: Maximum batch size for exports.
        endpoint: The endpoint to export logs to.
        headers: The headers to use for the export.
        certificate_file: The certificate file to use for the export.
        client_key_file: The client key file to use for the export.
        client_certificate_file: The client certificate file to use for the export.
        compression: The compression to use for the export.
    """

    service_name: str = Field(
        default="zenml",
        description="Name of the service for telemetry",
    )
    service_version: str = Field(
        default=__version__,
        description="Version of the service for telemetry",
    )
    max_queue_size: int = Field(
        default=LOGS_OTEL_MAX_QUEUE_SIZE,
        description="Maximum queue size for batch log processor",
    )
    schedule_delay_millis: int = Field(
        default=LOGS_OTEL_SCHEDULE_DELAY_MILLIS,
        description="Export interval in milliseconds",
    )
    max_export_batch_size: int = Field(
        default=LOGS_OTEL_MAX_EXPORT_BATCH_SIZE,
        description="Maximum batch size for exports",
    )
    export_timeout_millis: int = Field(
        default=LOGS_OTEL_EXPORT_TIMEOUT_MILLIS,
        description="Timeout for each export batch in milliseconds",
    )
    endpoint: str = Field(
        description="The endpoint to export logs to.",
    )
    headers: Optional[Dict[str, str]] = Field(
        default=None,
        description="The headers to use for the export.",
    )
    certificate_file: Optional[str] = Field(
        default=None,
        description="The certificate file to use for the export.",
    )
    client_key_file: Optional[str] = Field(
        default=None,
        description="The client key file to use for the export.",
    )
    client_certificate_file: Optional[str] = Field(
        default=None,
        description="The client certificate file to use for the export.",
    )
    compression: Compression = Field(
        default=Compression.NoCompression,
        description="The compression to use for the export.",
    )


class OtelLogStoreFlavor(Flavor):
    """OpenTelemetry log store flavor."""

    @property
    def name(self) -> str:
        """Name of the flavor.

        Returns:
            The name of the flavor.
        """
        return "otel"

    @property
    def docs_url(self) -> str:
        """URL to the flavor documentation.

        Returns:
            The URL to the flavor documentation.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> str:
        """URL to the SDK docs for this flavor.

        Returns:
            The URL to the SDK docs for this flavor.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """URL to the flavor logo.

        Returns:
            The URL to the flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/log_store/otel.png"

    @property
    def type(self) -> StackComponentType:
        """Stack component type.

        Returns:
            The stack component type.
        """
        return StackComponentType.LOG_STORE

    @property
    def config_class(self) -> Type[BaseLogStoreConfig]:
        """Returns `DatadogLogStoreConfig` config class.

        Returns:
            The config class.
        """
        return OtelLogStoreConfig

    @property
    def implementation_class(self) -> Type[BaseLogStore]:
        """Implementation class for this flavor.

        Returns:
            The implementation class.
        """
        from zenml.log_stores.otel.otel_log_store import OtelLogStore

        return OtelLogStore



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of Docker image builds to run ZenML pipelines."""

import itertools
import os
import subprocess
import sys
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
)

import zenml
from zenml.config.docker_settings import (
    DockerBuildConfig,
    DockerBuildOptions,
    DockerSettings,
    PythonEnvironmentExportMethod,
    PythonPackageInstaller,
)
from zenml.constants import (
    ENV_ZENML_CONFIG_PATH,
    ENV_ZENML_ENABLE_REPO_INIT_WARNINGS,
)
from zenml.enums import OperatingSystemType
from zenml.integrations.registry import integration_registry
from zenml.logger import get_logger
from zenml.utils import docker_utils, io_utils, source_utils

if TYPE_CHECKING:
    from zenml.code_repositories import BaseCodeRepository
    from zenml.container_registries import BaseContainerRegistry
    from zenml.image_builders import BuildContext
    from zenml.stack import Stack

logger = get_logger(__name__)

DOCKER_IMAGE_WORKDIR = "/app"
DOCKER_IMAGE_ZENML_CONFIG_PATH = f"{DOCKER_IMAGE_WORKDIR}/.zenconfig"

DEFAULT_DOCKER_PARENT_IMAGE = (
    f"zenmldocker/zenml:{zenml.__version__}-"
    f"py{sys.version_info.major}.{sys.version_info.minor}"
)
DEFAULT_ZENML_DOCKER_REPOSITORY = "zenml"

PIP_DEFAULT_ARGS = {
    "no-cache-dir": None,
    "default-timeout": 60,
}
UV_DEFAULT_ARGS = {"no-cache-dir": None}

DEFAULT_PYPROJECT_EXPORT_COMMANDS = [
    "uv export --format=requirements-txt --directory={directory} --no-hashes --no-emit-project",
    "poetry export --format=requirements.txt --directory={directory} --without-hashes",
]


class PipelineDockerImageBuilder:
    """Builds Docker images to run a ZenML pipeline."""

    def build_docker_image(
        self,
        docker_settings: "DockerSettings",
        tag: str,
        stack: "Stack",
        include_files: bool,
        entrypoint: Optional[str] = None,
        extra_files: Optional[Dict[str, str]] = None,
        code_repository: Optional["BaseCodeRepository"] = None,
        extra_requirements_files: Dict[str, List[str]] = {},
    ) -> Tuple[str, Optional[str], Optional[str]]:
        """Builds (and optionally pushes) a Docker image to run a pipeline.

        Use the image name returned by this method whenever you need to uniquely
        reference the pushed image in order to pull or run it.

        Args:
            docker_settings: The settings for the image build.
            tag: The tag to use for the image.
            stack: The stack on which the pipeline will be deployed.
            include_files: Whether to include files in the build context.
            entrypoint: Entrypoint to use for the final image. If left empty,
                no entrypoint will be included in the image.
            extra_files: Extra files to add to the build context. Keys are the
                path inside the build context, values are either the file
                content or a file path.
            code_repository: The code repository from which files will be
                downloaded.
            extra_requirements_files: Extra requirements to install in the
                Docker image. Each key is the name of a Python requirements file
                to be created and the value is the list of requirements to be
                installed.

        Returns:
            A tuple (image_digest, dockerfile, requirements):
            - The Docker image repo digest or local name, depending on whether
            the image was pushed or is just stored locally.
            - Dockerfile will contain the contents of the Dockerfile used to
            build the image.
            - Requirements is a string with a single pip requirement per line.

        Raises:
            RuntimeError: If the stack does not contain an image builder.
            ValueError: If no Dockerfile and/or custom parent image is
                specified and the Docker configuration doesn't require an
                image build.
            ValueError: If the specified Dockerfile does not exist.
        """
        requirements: Optional[str] = None
        dockerfile: Optional[str] = None

        if docker_settings.skip_build:
            assert (
                docker_settings.parent_image
            )  # checked via validator already

            # Should we tag this here and push it to the container registry of
            # the stack to make sure it's always accessible when running the
            # pipeline?
            return docker_settings.parent_image, dockerfile, requirements

        if docker_settings.dockerfile and not os.path.isfile(
            docker_settings.dockerfile
        ):
            raise ValueError(
                "Dockerfile at path "
                f"{os.path.abspath(docker_settings.dockerfile)} not found."
            )

        stack.validate()
        image_builder = stack.image_builder
        if not image_builder:
            raise RuntimeError(
                "Unable to build Docker images without an image builder in the "
                f"stack `{stack.name}`."
            )

        container_registry = stack.container_registry

        build_context_class = image_builder.build_context_class
        target_image_name = self._get_target_image_name(
            docker_settings=docker_settings,
            tag=tag,
            container_registry=container_registry,
        )

        requires_zenml_build = any(
            [
                docker_settings.requirements,
                docker_settings.pyproject_path,
                docker_settings.required_integrations,
                docker_settings.replicate_local_python_environment,
                docker_settings.install_stack_requirements,
                docker_settings.local_project_install_command,
                docker_settings.apt_packages,
                docker_settings.environment,
                include_files,
                entrypoint,
                extra_files,
                extra_requirements_files,
            ]
        )

        # Fallback to the value defined on the stack component if the
        # pipeline configuration doesn't have a configured value
        parent_image = (
            docker_settings.parent_image or DEFAULT_DOCKER_PARENT_IMAGE
        )

        if docker_settings.dockerfile:
            if parent_image != DEFAULT_DOCKER_PARENT_IMAGE:
                logger.warning(
                    "You've specified both a Dockerfile and a custom parent "
                    "image, ignoring the parent image."
                )

            push = (
                not image_builder.is_building_locally
                or not requires_zenml_build
            )

            if requires_zenml_build:
                # We will build an additional image on top of this one later
                # to include user files and/or install requirements. The image
                # we build now will be used as the parent for the next build.
                repository = docker_settings.target_repository
                if not repository:
                    if container_registry:
                        repository = (
                            container_registry.config.default_repository
                        )

                repository = repository or DEFAULT_ZENML_DOCKER_REPOSITORY
                user_image_tag = docker_utils.sanitize_tag(
                    f"{tag}-intermediate-build"
                )
                user_image_name = f"{repository}:{user_image_tag}"
                if push and container_registry:
                    user_image_name = (
                        f"{container_registry.config.uri}/{user_image_name}"
                    )

                parent_image = user_image_name
            else:
                # The image we'll build from the custom Dockerfile will be
                # used directly, so we tag it with the requested target name.
                user_image_name = target_image_name

            build_config = (
                docker_settings.parent_image_build_config
                or DockerBuildConfig()
            )
            build_context = build_context_class(
                root=docker_settings.build_context_root,
                dockerignore_file=build_config.dockerignore,
            )
            build_context.add_file(
                source=docker_settings.dockerfile, destination="Dockerfile"
            )
            logger.info("Building Docker image `%s`.", user_image_name)
            image_name_or_digest = image_builder.build(
                image_name=user_image_name,
                build_context=build_context,
                docker_build_options=build_config.build_options,
                container_registry=container_registry if push else None,
            )

        elif not requires_zenml_build:
            if parent_image == DEFAULT_DOCKER_PARENT_IMAGE:
                raise ValueError(
                    "Unable to run a ZenML pipeline with the given Docker "
                    "settings: No Dockerfile or custom parent image "
                    "specified and no files will be copied or requirements "
                    "installed."
                )
            else:
                # The parent image will be used directly to run the pipeline and
                # needs to be tagged/pushed
                docker_utils.tag_image(parent_image, target=target_image_name)
                if container_registry:
                    image_name_or_digest = container_registry.push_image(
                        target_image_name
                    )
                else:
                    image_name_or_digest = target_image_name

        if requires_zenml_build:
            logger.info("Building Docker image `%s`.", target_image_name)
            build_config = docker_settings.build_config or DockerBuildConfig()

            # Leave the build context empty if we don't want to include any files
            build_context_root = (
                source_utils.get_source_root() if include_files else None
            )
            dockerignore = (
                build_config.dockerignore or docker_settings.dockerignore
            )
            build_context = build_context_class(
                root=build_context_root,
                dockerignore_file=dockerignore,
            )

            requirements_files = self.gather_requirements_files(
                docker_settings=docker_settings,
                stack=stack,
                code_repository=code_repository,
                extra_requirements_files=extra_requirements_files,
            )

            self._add_requirements_files(
                requirements_files=requirements_files,
                build_context=build_context,
            )
            requirements = (
                "\n".join(
                    file_content for _, file_content, _ in requirements_files
                )
                or None
            )

            apt_packages = docker_settings.apt_packages.copy()
            if docker_settings.install_stack_requirements:
                apt_packages += stack.apt_packages

            # include apt packages from all required integrations
            for integration in docker_settings.required_integrations:
                # get the integration
                integration_cls = integration_registry.integrations[
                    integration
                ]
                apt_packages += integration_cls.APT_PACKAGES

            if apt_packages:
                logger.info(
                    "Including apt packages: %s",
                    ", ".join(f"`{p}`" for p in apt_packages),
                )

            if parent_image == DEFAULT_DOCKER_PARENT_IMAGE:
                # The default parent image is static and doesn't require a pull
                # each time
                pull_parent_image = False
            elif docker_settings.dockerfile and (
                not container_registry or image_builder.is_building_locally
            ):
                # We built a custom parent image and there was either no
                # container registry in the stack to push to or the image was
                # built locally and not pushed, so this is a local image
                pull_parent_image = False
            elif not image_builder.is_building_locally:
                # Remote image builders always need to pull the image
                pull_parent_image = True
            else:
                # If the image is local, we don't need to pull it. Otherwise
                # we play it safe and always pull in case the user pushed a new
                # image for the given name and tag
                pull_parent_image = not docker_utils.is_local_image(
                    parent_image
                )

            build_options = build_config.build_options or DockerBuildOptions()
            build_options.pull = pull_parent_image
            if build_options.rm is None:
                build_options.rm = False

            dockerfile = self._generate_zenml_pipeline_dockerfile(
                parent_image=parent_image,
                docker_settings=docker_settings,
                requirements_files=requirements_files,
                apt_packages=apt_packages,
                entrypoint=entrypoint,
            )
            build_context.add_file(destination="Dockerfile", source=dockerfile)

            if extra_files:
                for destination, source in extra_files.items():
                    build_context.add_file(
                        destination=destination, source=source
                    )

            image_name_or_digest = image_builder.build(
                image_name=target_image_name,
                build_context=build_context,
                docker_build_options=build_options,
                container_registry=container_registry,
            )

        return image_name_or_digest, dockerfile, requirements

    @staticmethod
    def _get_target_image_name(
        docker_settings: "DockerSettings",
        tag: str,
        container_registry: Optional["BaseContainerRegistry"] = None,
    ) -> str:
        """Returns the target image name.

        If a container registry is given, the image name will include the
        registry URI

        Args:
            docker_settings: The settings for the image build.
            tag: The tag to use for the image.
            container_registry: Optional container registry to which this
                image will be pushed.

        Returns:
            The docker image name.
        """
        repository = docker_settings.target_repository
        if not repository:
            if container_registry:
                repository = container_registry.config.default_repository

        repository = repository or DEFAULT_ZENML_DOCKER_REPOSITORY

        target_image_name = f"{repository}:{tag}"
        if container_registry:
            target_image_name = (
                f"{container_registry.config.uri}/{target_image_name}"
            )

        return target_image_name

    @classmethod
    def _add_requirements_files(
        cls,
        requirements_files: List[Tuple[str, str, List[str]]],
        build_context: "BuildContext",
    ) -> None:
        """Adds requirements files to the build context.

        Args:
            requirements_files: List of tuples
                (filename, file_content, pip_options).
            build_context: Build context to add the requirements files to.
        """
        for filename, file_content, _ in requirements_files:
            build_context.add_file(source=file_content, destination=filename)

    @staticmethod
    def gather_requirements_files(
        docker_settings: DockerSettings,
        stack: "Stack",
        code_repository: Optional["BaseCodeRepository"] = None,
        log: bool = True,
        extra_requirements_files: Dict[str, List[str]] = {},
    ) -> List[Tuple[str, str, List[str]]]:
        """Gathers and/or generates pip requirements files.

        This method is called in `PipelineDockerImageBuilder.build_docker_image`
        but it is also called by other parts of the codebase, e.g. the
        `AzureMLStepOperator`, which needs to upload the requirements files to
        AzureML where the step image is then built.

        Args:
            docker_settings: Docker settings that specifies which
                requirements to install.
            stack: The stack on which the pipeline will run.
            code_repository: The code repository from which files will be
                downloaded.
            log: If True, will log the requirements.
            extra_requirements_files: Extra requirements to install in the
                Docker image. Each key is the name of a Python requirements file
                to be created and the value is the list of requirements to be
                installed.

        Raises:
            RuntimeError: If the command to export the local python packages
                failed.
            FileNotFoundError: If the specified requirements file does not
                exist.

        Returns:
            List of tuples (filename, file_content, pip_options) of all
            requirements files.
            The files will be in the following order:
            - Packages installed in the local Python environment
            - Requirements defined by stack integrations
            - Extra requirements files
            - Requirements defined by user integrations
            - Requirements exported from a pyproject.toml
            - User-defined requirements
        """
        implicit_requirements = False
        pyproject_path = docker_settings.pyproject_path
        requirements = docker_settings.requirements

        if not any(
            [
                docker_settings.replicate_local_python_environment,
                requirements,
                pyproject_path,
            ]
        ):
            root = source_utils.get_source_root()
            requirements_path = os.path.join(root, "requirements.txt")
            pyproject_file_path = os.path.join(root, "pyproject.toml")

            if os.path.exists(requirements_path):
                implicit_requirements = True
                requirements = requirements_path
            elif os.path.exists(pyproject_file_path):
                implicit_requirements = True
                pyproject_path = pyproject_file_path

        if (
            implicit_requirements
            and docker_settings.disable_automatic_requirements_detection
        ):
            # TODO: This is only temporary to log a warning notifying users
            # that we will soon switch the default behavior to
            # `disable_automatic_requirements_detection=False`. Remove and
            # adjust the logic once we've made this change.
            if log:
                logger.warning(
                    "Detected `requirements.txt` or `pyproject.toml` files in "
                    "the source root. In future versions of ZenML, these will "
                    "be automatically picked up and installed in Docker images "
                    "by default. To disable this behavior and keep the "
                    "current behavior, set "
                    "`DockerSettings.disable_automatic_requirements_detection` "
                    "to `True`. If you want to enable this behavior right away, "
                    "you can do so by setting "
                    "`DockerSettings.disable_automatic_requirements_detection` "
                    "to `False`."
                )
            implicit_requirements = False
            requirements = None
            pyproject_path = None

        requirements_files: List[Tuple[str, str, List[str]]] = []

        # Generate requirements file for the local environment if configured
        if docker_settings.replicate_local_python_environment:
            if isinstance(
                docker_settings.replicate_local_python_environment,
                PythonEnvironmentExportMethod,
            ):
                command = (
                    docker_settings.replicate_local_python_environment.command
                )
            elif isinstance(
                docker_settings.replicate_local_python_environment, bool
            ):
                command = PythonEnvironmentExportMethod.PIP_FREEZE.command
            else:
                command = " ".join(
                    docker_settings.replicate_local_python_environment
                )

            try:
                local_requirements = subprocess.check_output(
                    command,
                    shell=True,  # nosec
                ).decode()
            except subprocess.CalledProcessError as e:
                raise RuntimeError(
                    "Unable to export local python packages."
                ) from e

            requirements_files.append(
                (".zenml_local_requirements", local_requirements, [])
            )
            if log:
                logger.info(
                    "- Including python packages from local environment"
                )

        if docker_settings.install_stack_requirements:
            stack_requirements = stack.requirements()
            if code_repository:
                stack_requirements.update(code_repository.requirements)

            if stack_requirements:
                stack_requirements_list = sorted(stack_requirements)
                stack_requirements_file = "\n".join(stack_requirements_list)
                requirements_files.append(
                    (
                        ".zenml_stack_integration_requirements",
                        stack_requirements_file,
                        [],
                    )
                )
                if log:
                    logger.info(
                        "- Including stack requirements: %s",
                        ", ".join(f"`{r}`" for r in stack_requirements_list),
                    )

        for filename, extra_requirements in extra_requirements_files.items():
            requirements_list = sorted(extra_requirements)
            requirements_file = "\n".join(requirements_list)
            requirements_files.append((filename, requirements_file, []))
            if log:
                logger.info(
                    "- Including extra requirements from file `%s`: %s",
                    filename,
                    ", ".join(f"`{r}`" for r in requirements_list),
                )

        # Generate requirements file for all required integrations
        integration_requirements = set(
            itertools.chain.from_iterable(
                integration_registry.select_integration_requirements(
                    integration_name=integration,
                    target_os=OperatingSystemType.LINUX,
                )
                for integration in docker_settings.required_integrations
            )
        )

        if integration_requirements:
            integration_requirements_list = sorted(integration_requirements)
            integration_requirements_file = "\n".join(
                integration_requirements_list
            )
            requirements_files.append(
                (
                    ".zenml_integration_requirements",
                    integration_requirements_file,
                    [],
                )
            )
            if log:
                logger.info(
                    "- Including integration requirements: %s",
                    ", ".join(f"`{r}`" for r in integration_requirements_list),
                )

        if pyproject_path:
            path = os.path.abspath(pyproject_path)

            if not os.path.exists(path):
                raise FileNotFoundError(
                    f"Pyproject file {path} does not exist."
                )

            def _run_command(command: str) -> str:
                command = command.format(directory=os.path.dirname(path))

                result = subprocess.run(
                    command,
                    capture_output=True,
                    check=True,
                    shell=True,  # nosec
                )
                return result.stdout.decode()

            if docker_settings.pyproject_export_command:
                command = " ".join(docker_settings.pyproject_export_command)

                try:
                    pyproject_requirements = _run_command(command)
                except subprocess.CalledProcessError as e:
                    raise RuntimeError(
                        "Failed to export pyproject dependencies with "
                        f"command `{command}`: {e.stderr.decode()}"
                    )
            else:
                command_errors = {}

                for command in DEFAULT_PYPROJECT_EXPORT_COMMANDS:
                    try:
                        pyproject_requirements = _run_command(command)
                    except subprocess.CalledProcessError as e:
                        command_errors[command] = e.stderr.decode()
                    else:
                        break
                else:
                    raise RuntimeError(
                        "Failed to export pyproject dependencies with the "
                        f"following commands: {command_errors}. Please specify "
                        "a working command to export your pyproject.toml "
                        "dependencies to a requirements.txt formatted file "
                        "using `DockerSettings.pyproject_export_command`."
                    )

            requirements_files.append(
                (".zenml_pyproject_requirements", pyproject_requirements, [])
            )
            if log:
                logger.info(
                    "- %s python packages from file `%s`",
                    "Implicitly including"
                    if implicit_requirements
                    else "Including",
                    path,
                )

        # Generate/Read requirements file for user-defined requirements
        if isinstance(requirements, str):
            path = os.path.abspath(requirements)
            try:
                user_requirements = io_utils.read_file_contents_as_string(path)
            except FileNotFoundError as e:
                raise FileNotFoundError(
                    f"Requirements file {path} does not exist."
                ) from e
            if log:
                logger.info(
                    "- %s user-defined requirements from file `%s`",
                    "Implicitly including"
                    if implicit_requirements
                    else "Including",
                    path,
                )
        elif isinstance(requirements, List):
            user_requirements = "\n".join(requirements)
            if log:
                logger.info(
                    "- Including user-defined requirements: %s",
                    ", ".join(f"`{r}`" for r in requirements),
                )
        else:
            user_requirements = None

        if user_requirements:
            requirements_files.append(
                (".zenml_user_requirements", user_requirements, [])
            )

        return requirements_files

    @staticmethod
    def _generate_zenml_pipeline_dockerfile(
        parent_image: str,
        docker_settings: DockerSettings,
        requirements_files: Sequence[Tuple[str, str, List[str]]] = (),
        apt_packages: Sequence[str] = (),
        entrypoint: Optional[str] = None,
    ) -> str:
        """Generates a Dockerfile.

        Args:
            parent_image: The image to use as parent for the Dockerfile.
            docker_settings: Docker settings for this image build.
            requirements_files: List of tuples that contain three items:
                - the name of a requirements file,
                - the content of that file,
                - options that should be passed to pip when installing the
                    requirements file.
            apt_packages: APT packages to install.
            entrypoint: The default entrypoint command that gets executed when
                running a container of an image created by this Dockerfile.

        Raises:
            ValueError: If an unsupported python package installer was
                configured.

        Returns:
            The generated Dockerfile.
        """
        lines = [f"FROM {parent_image}", f"WORKDIR {DOCKER_IMAGE_WORKDIR}"]

        if (
            docker_settings.build_config
            and docker_settings.build_config.build_options
        ):
            b_args = docker_settings.build_config.build_options.build_args
            if b_args:
                for arg_key in b_args.keys():
                    lines.append(f"ARG {arg_key}")

        for key, value in docker_settings.environment.items():
            lines.append(f"ENV {key.upper()}='{value}'")

        if apt_packages:
            apt_packages = " ".join(f"'{p}'" for p in apt_packages)

            lines.append(
                "RUN apt-get update && apt-get install -y "
                f"--no-install-recommends {apt_packages}"
            )

        if (
            docker_settings.python_package_installer
            == PythonPackageInstaller.PIP
        ):
            install_command = "pip install"
            default_installer_args: Dict[str, Any] = PIP_DEFAULT_ARGS
        elif (
            docker_settings.python_package_installer
            == PythonPackageInstaller.UV
        ):
            lines.append("RUN pip install uv")
            install_command = "uv pip install"
            default_installer_args = UV_DEFAULT_ARGS
        else:
            raise ValueError("Unsupported python package installer.")

        installer_args = {
            **default_installer_args,
            **docker_settings.python_package_installer_args,
        }
        installer_args_string = " ".join(
            f"--{key}" if value is None else f"--{key}={value}"
            for key, value in installer_args.items()
        )
        for file, _, options in requirements_files:
            lines.append(f"COPY {file} .")
            option_string = " ".join(options)

            lines.append(
                f"RUN {install_command} {installer_args_string}"
                f"{option_string} -r {file}"
            )

        lines.append(f"ENV {ENV_ZENML_ENABLE_REPO_INIT_WARNINGS}=False")
        lines.append(
            f"ENV {ENV_ZENML_CONFIG_PATH}={DOCKER_IMAGE_ZENML_CONFIG_PATH}"
        )

        lines.append("COPY . .")
        lines.append("RUN chmod -R a+rw .")

        if docker_settings.local_project_install_command:
            lines.append(
                f"RUN {docker_settings.local_project_install_command}"
            )

        if docker_settings.user:
            # Change file ownership to specified user
            lines.append(f"RUN chown -R {docker_settings.user} .")
            # Switch back to specified user for subsequent instructions
            lines.append(f"USER {docker_settings.user}")

        for key, value in docker_settings.runtime_environment.items():
            lines.append(f"ENV {key.upper()}='{value}'")

        if entrypoint:
            lines.append(f"ENTRYPOINT {entrypoint}")

        return "\n".join(lines)



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Utility functions to handle metadata for ZenML entities."""

from typing import Dict, List, Optional, Set, Union, overload
from uuid import UUID

from zenml.client import Client
from zenml.enums import MetadataResourceTypes, ModelStages
from zenml.logger import get_logger
from zenml.metadata.metadata_types import MetadataType
from zenml.models import (
    ArtifactVersionIdentifier,
    ModelVersionIdentifier,
    PipelineRunIdentifier,
    RunMetadataResource,
    StepRunIdentifier,
)
from zenml.steps.step_context import get_step_context

logger = get_logger(__name__)


@overload
def log_metadata(
    metadata: Dict[str, MetadataType],
) -> None: ...


@overload
def log_metadata(
    *,
    metadata: Dict[str, MetadataType],
    step_id: UUID,
) -> None: ...


@overload
def log_metadata(
    *,
    metadata: Dict[str, MetadataType],
    step_name: str,
    run_id_name_or_prefix: Union[UUID, str],
) -> None: ...


@overload
def log_metadata(
    *,
    metadata: Dict[str, MetadataType],
    run_id_name_or_prefix: Union[UUID, str],
) -> None: ...


@overload
def log_metadata(
    *,
    metadata: Dict[str, MetadataType],
    artifact_version_id: UUID,
) -> None: ...


@overload
def log_metadata(
    *,
    metadata: Dict[str, MetadataType],
    artifact_name: str,
    artifact_version: Optional[str] = None,
) -> None: ...


@overload
def log_metadata(
    *,
    metadata: Dict[str, MetadataType],
    infer_artifact: bool = False,
    artifact_name: Optional[str] = None,
) -> None: ...


# Model Metadata
@overload
def log_metadata(
    *,
    metadata: Dict[str, MetadataType],
    model_version_id: UUID,
) -> None: ...


@overload
def log_metadata(
    *,
    metadata: Dict[str, MetadataType],
    model_name: str,
    model_version: Union[ModelStages, int, str],
) -> None: ...


@overload
def log_metadata(
    *,
    metadata: Dict[str, MetadataType],
    infer_model: bool = False,
) -> None: ...


def log_metadata(
    metadata: Dict[str, MetadataType],
    # Steps and runs
    step_id: Optional[UUID] = None,
    step_name: Optional[str] = None,
    run_id_name_or_prefix: Optional[Union[UUID, str]] = None,
    # Artifacts
    artifact_version_id: Optional[UUID] = None,
    artifact_name: Optional[str] = None,
    artifact_version: Optional[str] = None,
    infer_artifact: bool = False,
    # Models
    model_version_id: Optional[UUID] = None,
    model_name: Optional[str] = None,
    model_version: Optional[Union[ModelStages, int, str]] = None,
    infer_model: bool = False,
) -> None:
    """Logs metadata for various resource types in a generalized way.

    Args:
        metadata: The metadata to log.
        step_id: The ID of the step.
        step_name: The name of the step.
        run_id_name_or_prefix: The id, name or prefix of the run
        artifact_version_id: The ID of the artifact version
        artifact_name: The name of the artifact.
        artifact_version: The version of the artifact.
        infer_artifact: Flag deciding whether the artifact version should be
            inferred from the step context.
        model_version_id: The ID of the model version.
        model_name: The name of the model.
        model_version: The version of the model.
        infer_model: Flag deciding whether the model version should be
            inferred from the step context.

    Raises:
        ValueError: If no identifiers are provided and the function is not
            called from within a step.
    """
    client = Client()

    resources: List[RunMetadataResource] = []
    publisher_step_id = None

    # Log metadata to a step by ID
    if step_id is not None:
        resources = [
            RunMetadataResource(
                id=step_id, type=MetadataResourceTypes.STEP_RUN
            )
        ]

    # Log metadata to a step by name and run ID
    elif step_name is not None and run_id_name_or_prefix is not None:
        run = client.get_pipeline_run(name_id_or_prefix=run_id_name_or_prefix)
        step_model_id = client.list_run_steps(
            pipeline_run_id=run.id, name=step_name
        )[0].id
        resources = [
            RunMetadataResource(
                id=step_model_id, type=MetadataResourceTypes.STEP_RUN
            )
        ]

    # Log metadata to a run by ID
    elif run_id_name_or_prefix is not None:
        run_model = client.get_pipeline_run(
            name_id_or_prefix=run_id_name_or_prefix
        )
        resources = [
            RunMetadataResource(
                id=run_model.id, type=MetadataResourceTypes.PIPELINE_RUN
            )
        ]

    # Log metadata to a model version by name and version
    elif model_name is not None and model_version is not None:
        model_version_model = client.get_model_version(
            model_name_or_id=model_name,
            model_version_name_or_number_or_id=model_version,
        )
        resources = [
            RunMetadataResource(
                id=model_version_model.id,
                type=MetadataResourceTypes.MODEL_VERSION,
            )
        ]

    # Log metadata to a model version by id
    elif model_version_id is not None:
        resources = [
            RunMetadataResource(
                id=model_version_id,
                type=MetadataResourceTypes.MODEL_VERSION,
            )
        ]

    # Log metadata to a model through the step context
    elif infer_model is True:
        try:
            step_context = get_step_context()
        except RuntimeError:
            raise ValueError(
                "If you are using the `infer_model` option, the function must "
                "be called inside a step with configured `model` in decorator."
                "Otherwise, you can provide a `model_version_id` or a "
                "combination of `model_name` and `model_version`."
            )

        if step_context.model_version is None:
            raise ValueError(
                "The step context does not feature any model versions."
            )

        resources = [
            RunMetadataResource(
                id=step_context.model_version.id,
                type=MetadataResourceTypes.MODEL_VERSION,
            )
        ]

    # Log metadata to an artifact version by its name and version
    elif artifact_name is not None and artifact_version is not None:
        artifact_version_model = client.get_artifact_version(
            name_id_or_prefix=artifact_name, version=artifact_version
        )
        resources = [
            RunMetadataResource(
                id=artifact_version_model.id,
                type=MetadataResourceTypes.ARTIFACT_VERSION,
            )
        ]

    # Log metadata to an artifact version by its ID
    elif artifact_version_id is not None:
        resources = [
            RunMetadataResource(
                id=artifact_version_id,
                type=MetadataResourceTypes.ARTIFACT_VERSION,
            )
        ]

    # Log metadata to an artifact version through the step context
    elif infer_artifact is True:
        try:
            step_context = get_step_context()
        except RuntimeError:
            raise ValueError(
                "When you are using the `infer_artifact` option when you call "
                "`log_metadata`, it must be called inside a step with outputs."
                "Otherwise, you can provide a `artifact_version_id` or a "
                "combination of `artifact_name` and `artifact_version`."
            )

        step_output_names = list(step_context._outputs.keys())

        if artifact_name is not None:
            # If a name provided, ensure it is in the outputs
            if artifact_name not in step_output_names:
                raise ValueError(
                    f"The provided artifact name`{artifact_name}` does not "
                    f"exist in the step outputs: {step_output_names}."
                )
        else:
            # If no name provided, ensure there is only one output
            if len(step_output_names) > 1:
                raise ValueError(
                    "There is more than one output. If you would like to use "
                    "the `infer_artifact` option, you need to define an "
                    "`artifact_name`."
                )

            if len(step_output_names) == 0:
                raise ValueError("The step does not have any outputs.")

            artifact_name = step_output_names[0]

        step_context.add_output_metadata(
            metadata=metadata, output_name=artifact_name
        )
        return

    # If every additional value is None, that means we are calling it bare bones
    # and this call needs to happen during a step execution. We will use the
    # step context to fetch the step, run and possibly the model version and
    # attach the metadata accordingly.
    elif all(
        v is None
        for v in [
            step_id,
            step_name,
            run_id_name_or_prefix,
            artifact_version_id,
            artifact_name,
            artifact_version,
            model_version_id,
            model_name,
            model_version,
        ]
    ):
        try:
            step_context = get_step_context()
        except RuntimeError:
            raise ValueError(
                "You are calling 'log_metadata()' outside of a step execution. "
                "If you would like to add metadata to a ZenML entity outside "
                "of the step execution, please provide the required "
                "identifiers."
            )

        resources = [
            RunMetadataResource(
                id=step_context.step_run.id,
                type=MetadataResourceTypes.STEP_RUN,
            )
        ]
        publisher_step_id = step_context.step_run.id

    else:
        raise ValueError(
            """
            Unsupported way to call the `log_metadata`. Possible combinations "
            include:
            
            # Automatic logging to a step (within a step)
            log_metadata(metadata={})
            
            # Manual logging to a step
            log_metadata(metadata={}, step_name=..., run_id_name_or_prefix=...)
            log_metadata(metadata={}, step_id=...)
            
            # Manual logging to a run
            log_metadata(metadata={}, run_id_name_or_prefix=...)
            
            # Automatic logging to a model (within a step)
            log_metadata(metadata={}, infer_model=True)
            
            # Manual logging to a model
            log_metadata(metadata={}, model_name=..., model_version=...)
            log_metadata(metadata={}, model_version_id=...)
            
            # Automatic logging to an artifact (within a step)
            log_metadata(metadata={}, infer_artifact=True)  # step with single output
            log_metadata(metadata={}, artifact_name=..., infer_artifact=True)  # specific output of a step
            
            # Manual logging to an artifact
            log_metadata(metadata={}, artifact_name=..., artifact_version=...)
            log_metadata(metadata={}, artifact_version_id=...)
            """
        )

    client.create_run_metadata(
        metadata=metadata,
        resources=resources,
        publisher_step_id=publisher_step_id,
    )


def bulk_log_metadata(
    metadata: Dict[str, MetadataType],
    pipeline_runs: list[PipelineRunIdentifier] | None = None,
    step_runs: list[StepRunIdentifier] | None = None,
    artifact_versions: list[ArtifactVersionIdentifier] | None = None,
    model_versions: list[ModelVersionIdentifier] | None = None,
    infer_models: bool = False,
    infer_artifacts: bool = False,
) -> None:
    """Logs metadata for multiple entities in a single invocation.

    Args:
        metadata: The metadata to log.
        pipeline_runs: A list of pipeline runs to log metadata for.
        step_runs: A list of step runs to log metadata for.
        artifact_versions: A list of artifact versions to log metadata for.
        model_versions: A list of model versions to log metadata for.
        infer_models: Flag - when enabled infer model to log metadata for from step context.
        infer_artifacts: Flag - when enabled infer artifact to log metadata for from step context.

    Raises:
        ValueError: If options are not passed correctly (empty metadata or no identifier options) or
            invocation with `infer` options is done outside of a step context.
    """
    client = Client()

    resources: Set[RunMetadataResource] = set()

    if not metadata:
        raise ValueError("You must provide metadata to log.")

    if not any(
        bool(v)
        for v in [
            pipeline_runs,
            step_runs,
            artifact_versions,
            model_versions,
            infer_models,
            infer_artifacts,
        ]
    ):
        raise ValueError(
            "You must select at least one entity to log metadata to."
        )

    try:
        step_context = get_step_context()
    except RuntimeError:
        step_context = None

    if (infer_models or infer_artifacts) and step_context is None:
        raise ValueError(
            "Infer options can be used only within a step function code."
        )

    # resolve pipeline runs and add metadata resources

    for run in pipeline_runs or []:
        if not run.id:
            run.id = client.get_pipeline_run(name_id_or_prefix=run.value).id
        resources.add(
            RunMetadataResource(
                id=run.id, type=MetadataResourceTypes.PIPELINE_RUN
            )
        )

    # resolve step runs and add metadata resources

    for step in step_runs or []:
        if not step.id and (step.name and step.run):
            run_model = client.get_pipeline_run(
                name_id_or_prefix=step.run.value
            )
            step.id = client.list_run_steps(
                pipeline_run_id=run_model.id, name=step.name
            )[0].id

        resources.add(
            RunMetadataResource(
                id=step.id, type=MetadataResourceTypes.STEP_RUN
            )
        )

    # resolve artifacts and add metadata resources

    for artifact_version in artifact_versions or []:
        if not artifact_version.id and (
            artifact_version.name and artifact_version.version
        ):
            artifact_version.id = client.get_artifact_version(
                name_id_or_prefix=artifact_version.name,
                version=artifact_version.version,
            ).id
        resources.add(
            RunMetadataResource(
                id=artifact_version.id,
                type=MetadataResourceTypes.ARTIFACT_VERSION,
            )
        )

    # resolve models and add metadata resources

    for model_version in model_versions or []:
        if not model_version.id:
            model_version.id = client.get_model_version(
                model_name_or_id=model_version.name,
                model_version_name_or_number_or_id=model_version.version,
            ).id
        resources.add(
            RunMetadataResource(
                id=model_version.id, type=MetadataResourceTypes.MODEL_VERSION
            )
        )

    # infer models - resolve from step context

    if infer_models and step_context and not step_context.model_version:
        raise ValueError(
            "The step context does not feature any model versions."
        )
    elif infer_models and step_context and step_context.model_version:
        resources.add(
            RunMetadataResource(
                id=step_context.model_version.id,
                type=MetadataResourceTypes.MODEL_VERSION,
            )
        )

    # infer artifacts - resolve from step context

    if infer_artifacts and step_context:
        step_output_names = list(step_context._outputs.keys())

        for artifact_name in step_output_names:
            step_context.add_output_metadata(
                metadata=metadata, output_name=artifact_name
            )

    if not resources:
        return

    client.create_run_metadata(
        metadata=metadata,
        resources=list(resources),
        publisher_step_id=None,
    )



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Huggingface deployer flavor."""

from typing import TYPE_CHECKING, Optional, Type

from pydantic import Field

from zenml.deployers.base_deployer import (
    BaseDeployerConfig,
    BaseDeployerFlavor,
    BaseDeployerSettings,
)
from zenml.integrations.huggingface import HUGGINGFACE_DEPLOYER_FLAVOR
from zenml.models import ServiceConnectorRequirements
from zenml.utils.secret_utils import SecretField

if TYPE_CHECKING:
    from zenml.integrations.huggingface.deployers import HuggingFaceDeployer


class HuggingFaceDeployerSettings(BaseDeployerSettings):
    """Hugging Face deployer settings.

    Attributes:
        space_hardware: Hardware tier for the Space (e.g., 'cpu-basic', 't4-small')
        space_storage: Persistent storage tier (e.g., 'small', 'medium', 'large')
        private: Whether to create a private Space (default: True for security)
    """

    space_hardware: Optional[str] = Field(
        default=None,
        description="Hardware tier for Space execution. Controls compute resources "
        "available to the deployed pipeline. Options: 'cpu-basic' (2 vCPU, 16GB RAM), "
        "'cpu-upgrade' (8 vCPU, 32GB RAM), 't4-small' (4 vCPU, 15GB RAM, NVIDIA T4), "
        "'t4-medium' (8 vCPU, 30GB RAM, NVIDIA T4). See "
        "https://huggingface.co/docs/hub/spaces-gpus for full list. Defaults to "
        "cpu-basic if not specified",
    )
    space_storage: Optional[str] = Field(
        default=None,
        description="Persistent storage tier for Space data. Determines available disk "
        "space for artifacts and logs. Options: 'small' (20GB), 'medium' (150GB), "
        "'large' (1TB). Storage persists across Space restarts. If not specified, "
        "uses ephemeral storage that resets on restart",
    )
    private: bool = Field(
        default=True,
        description="Controls whether the deployed Space is private or public. "
        "Private Spaces are only accessible to the owner and authorized users. "
        "Public Spaces are visible to anyone with the URL. Defaults to True for security",
    )


class HuggingFaceDeployerConfig(
    BaseDeployerConfig, HuggingFaceDeployerSettings
):
    """Configuration for the Hugging Face deployer."""

    token: Optional[str] = SecretField(
        default=None,
        description="Hugging Face API token for authentication with write permissions. "
        "Can reference a ZenML secret using {{secret_name.key}} syntax or provide "
        "the token directly. Create tokens at https://huggingface.co/settings/tokens "
        "with 'write' access enabled. Example: '{{hf_token.token}}' references the "
        "'token' key in the 'hf_token' secret",
    )
    organization: Optional[str] = Field(
        default=None,
        description="Hugging Face organization name to deploy Spaces under. If not "
        "specified, Spaces are created under the authenticated user's account. "
        "Example: 'zenml' deploys to https://huggingface.co/spaces/zenml/. "
        "Requires organization membership with appropriate permissions",
    )
    space_prefix: str = Field(
        default="zenml",
        description="Prefix for Space names to organize deployments and avoid naming "
        "conflicts. Combined with deployment name to form the full Space ID. "
        "Example: prefix 'zenml' with deployment 'my-pipeline' creates Space "
        "'zenml-my-pipeline'. Maximum combined length is 96 characters",
    )


class HuggingFaceDeployerFlavor(BaseDeployerFlavor):
    """Flavor for the Hugging Face deployer."""

    @property
    def name(self) -> str:
        """Name of the flavor.

        Returns:
            The flavor name.
        """
        return HUGGINGFACE_DEPLOYER_FLAVOR

    @property
    def docs_url(self) -> Optional[str]:
        """A URL to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A URL to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A URL to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/deployer/huggingface.png"

    @property
    def config_class(self) -> Type[BaseDeployerConfig]:
        """Returns `HuggingFaceDeployerConfig` config class.

        Returns:
            The config class.
        """
        return HuggingFaceDeployerConfig

    @property
    def implementation_class(self) -> Type["HuggingFaceDeployer"]:
        """Implementation class for this flavor.

        Returns:
            The implementation class.
        """
        from zenml.integrations.huggingface.deployers import (
            HuggingFaceDeployer,
        )

        return HuggingFaceDeployer

    @property
    def service_connector_requirements(
        self,
    ) -> Optional[ServiceConnectorRequirements]:
        """Service connector resource requirements for this flavor.

        Returns:
            Service connector resource requirements.
        """
        return None



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Models representing curated visualizations."""

from typing import TYPE_CHECKING, Optional
from uuid import UUID

from pydantic import Field, NonNegativeInt

from zenml.enums import CuratedVisualizationSize, VisualizationResourceTypes
from zenml.models.v2.base.base import BaseUpdate
from zenml.models.v2.base.scoped import (
    ProjectScopedRequest,
    ProjectScopedResponse,
    ProjectScopedResponseBody,
    ProjectScopedResponseMetadata,
    ProjectScopedResponseResources,
)

if TYPE_CHECKING:
    from zenml.models.v2.core.artifact_visualization import (
        ArtifactVisualizationResponse,
    )


# ------------------ Request Model ------------------


class CuratedVisualizationRequest(ProjectScopedRequest):
    """Request model for curated visualizations.

    Each curated visualization links a pre-rendered artifact visualization
    to a single ZenML resource to surface it in the appropriate UI context.
    Supported resources include:
    - **Deployments**
    - **Models**
    - **Pipelines**
    - **Pipeline Runs**
    - **Pipeline Snapshots**
    - **Projects**
    """

    artifact_visualization_id: UUID = Field(
        title="The artifact visualization ID.",
        description=(
            "Identifier of the artifact visualization that should be surfaced "
            "for the target resource."
        ),
    )
    display_name: Optional[str] = Field(
        default=None,
        title="The display name of the visualization.",
    )
    display_order: Optional[NonNegativeInt] = Field(
        default=None,
        title="The display order of the visualization.",
        description=(
            "Optional ordering hint that must be unique for the combination "
            "of resource type and resource ID."
        ),
    )
    layout_size: CuratedVisualizationSize = Field(
        default=CuratedVisualizationSize.FULL_WIDTH,
        title="The layout size of the visualization.",
        description=(
            "Controls how much horizontal space the visualization occupies "
            "on the dashboard."
        ),
    )
    resource_id: UUID = Field(
        title="The linked resource ID.",
        description=(
            "Identifier of the resource (deployment, model, pipeline, pipeline "
            "run, pipeline snapshot, or project) that should surface this "
            "visualization."
        ),
    )
    resource_type: VisualizationResourceTypes = Field(
        title="The linked resource type.",
        description="Type of the resource associated with this visualization.",
    )


# ------------------ Update Model ------------------


class CuratedVisualizationUpdate(BaseUpdate):
    """Update model for curated visualizations."""

    display_name: Optional[str] = Field(
        default=None,
        title="The new display name of the visualization.",
    )
    display_order: Optional[NonNegativeInt] = Field(
        default=None,
        title="The new display order of the visualization.",
        description=(
            "Optional ordering hint. When provided, it must remain unique for "
            "the combination of resource type and resource ID."
        ),
    )
    layout_size: Optional[CuratedVisualizationSize] = Field(
        default=None,
        title="The updated layout size of the visualization.",
    )


# ------------------ Response Model ------------------


class CuratedVisualizationResponseBody(ProjectScopedResponseBody):
    """Response body for curated visualizations."""

    artifact_visualization_id: UUID = Field(
        title="The artifact visualization ID.",
        description=(
            "Identifier of the artifact visualization that is curated for this resource."
        ),
    )
    artifact_version_id: UUID = Field(
        title="The artifact version ID.",
        description=(
            "Identifier of the artifact version that owns the curated visualization. "
            "Provided for read-only context when available."
        ),
    )
    display_name: Optional[str] = Field(
        default=None,
        title="The display name of the visualization.",
    )
    display_order: Optional[NonNegativeInt] = Field(
        default=None,
        title="The display order of the visualization.",
        description=(
            "Optional ordering hint that is unique per combination of "
            "resource type and resource ID."
        ),
    )
    layout_size: CuratedVisualizationSize = Field(
        default=CuratedVisualizationSize.FULL_WIDTH,
        title="The layout size of the visualization.",
    )
    resource_id: UUID = Field(
        title="The linked resource ID.",
        description="Identifier of the resource associated with this visualization.",
    )
    resource_type: VisualizationResourceTypes = Field(
        title="The linked resource type.",
        description="Type of the resource associated with this visualization.",
    )


class CuratedVisualizationResponseMetadata(ProjectScopedResponseMetadata):
    """Response metadata for curated visualizations."""


class CuratedVisualizationResponseResources(ProjectScopedResponseResources):
    """Response resources included for curated visualizations."""

    artifact_visualization: "ArtifactVisualizationResponse" = Field(
        title="The artifact visualization.",
        description=(
            "Artifact visualization that is surfaced through this curated visualization."
        ),
    )


class CuratedVisualizationResponse(
    ProjectScopedResponse[
        CuratedVisualizationResponseBody,
        CuratedVisualizationResponseMetadata,
        CuratedVisualizationResponseResources,
    ]
):
    """Response model for curated visualizations."""

    def get_hydrated_version(self) -> "CuratedVisualizationResponse":
        """Get the hydrated version of this curated visualization.

        Returns:
            A hydrated instance of the same entity.
        """
        from zenml.client import Client

        client = Client()
        return client.zen_store.get_curated_visualization(self.id)

    # Helper properties
    @property
    def artifact_visualization_id(self) -> UUID:
        """The artifact visualization ID.

        Returns:
            The artifact visualization ID.
        """
        return self.get_body().artifact_visualization_id

    @property
    def artifact_version_id(self) -> UUID:
        """The artifact version ID.

        Returns:
            The artifact version ID if available.
        """
        return self.get_body().artifact_version_id

    @property
    def display_name(self) -> Optional[str]:
        """The display name of the visualization.

        Returns:
            The display name of the visualization.
        """
        return self.get_body().display_name

    @property
    def display_order(self) -> Optional[int]:
        """The display order of the visualization.

        Returns:
            The display order of the visualization.
        """
        return self.get_body().display_order

    @property
    def layout_size(self) -> CuratedVisualizationSize:
        """The layout size of the visualization.

        Returns:
            The layout size of the visualization.
        """
        return self.get_body().layout_size

    @property
    def artifact_visualization(self) -> "ArtifactVisualizationResponse":
        """The curated artifact visualization resource.

        Returns:
            The artifact visualization resource.
        """
        return self.get_resources().artifact_visualization

    @property
    def resource_id(self) -> UUID:
        """The identifier of the linked resource.

        Returns:
            The resource identifier associated with this visualization.
        """
        return self.get_body().resource_id

    @property
    def resource_type(self) -> VisualizationResourceTypes:
        """The type of the linked resource.

        Returns:
            The resource type associated with this visualization.
        """
        return self.get_body().resource_type



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""RBAC SQL Zen Store implementation."""

from typing import (
    Optional,
    Tuple,
)
from uuid import UUID

from zenml.logger import get_logger
from zenml.models import (
    ModelRequest,
    ModelResponse,
    ModelVersionRequest,
    ModelVersionResponse,
)
from zenml.zen_server.feature_gate.endpoint_utils import (
    check_entitlement,
    report_usage,
)
from zenml.zen_server.rbac.models import Action, ResourceType
from zenml.zen_server.rbac.utils import (
    verify_permission,
    verify_permission_for_model,
)
from zenml.zen_stores.sql_zen_store import SqlZenStore

logger = get_logger(__name__)


class RBACSqlZenStore(SqlZenStore):
    """Wrapper around the SQLZenStore that implements RBAC functionality."""

    def _get_or_create_model(
        self, model_request: ModelRequest
    ) -> Tuple[bool, ModelResponse]:
        """Get or create a model.

        Args:
            model_request: The model request.

        # noqa: DAR401
        Raises:
            Exception: If the user is not allowed to create a model.

        Returns:
            A boolean whether the model was created or not, and the model.
        """
        allow_model_creation = True
        error = None

        try:
            verify_permission(
                resource_type=ResourceType.MODEL,
                action=Action.CREATE,
                project_id=model_request.project,
            )
            check_entitlement(feature=ResourceType.MODEL)
        except Exception as e:
            allow_model_creation = False
            error = e

        if allow_model_creation:
            created, model_response = super()._get_or_create_model(
                model_request
            )
        else:
            try:
                model_response = self.get_model_by_name_or_id(
                    model_name_or_id=model_request.name,
                    project=model_request.project,
                )
                created = False
            except KeyError:
                # The model does not exist. We now raise the error that
                # explains why the model could not be created, instead of just
                # the KeyError that it doesn't exist
                assert error
                raise error from None

        if created:
            report_usage(
                feature=ResourceType.MODEL, resource_id=model_response.id
            )
        else:
            verify_permission_for_model(model_response, action=Action.READ)

        return created, model_response

    def _get_model_version(
        self,
        model_id: UUID,
        version_name: Optional[str] = None,
        producer_run_id: Optional[UUID] = None,
    ) -> ModelVersionResponse:
        """Get a model version.

        Args:
            model_id: The ID of the model.
            version_name: The name of the model version.
            producer_run_id: The ID of the producer pipeline run. If this is
                set, only numeric versions created as part of the pipeline run
                will be returned.

        Returns:
            The model version.
        """
        model_version = super()._get_model_version(
            model_id=model_id,
            version_name=version_name,
            producer_run_id=producer_run_id,
        )
        verify_permission_for_model(model_version, action=Action.READ)
        return model_version

    def _get_or_create_model_version(
        self,
        model_version_request: ModelVersionRequest,
        producer_run_id: Optional[UUID] = None,
    ) -> Tuple[bool, ModelVersionResponse]:
        """Get or create a model version.

        Args:
            model_version_request: The model version request.
            producer_run_id: ID of the producer pipeline run.

        # noqa: DAR401
        Raises:
            Exception: If the authenticated user is not allowed to
                create a model version.

        Returns:
            A boolean whether the model version was created or not, and the
            model version.
        """
        allow_creation = True
        error = None

        try:
            verify_permission(
                resource_type=ResourceType.MODEL_VERSION,
                action=Action.CREATE,
                project_id=model_version_request.project,
            )
        except Exception as e:
            allow_creation = False
            error = e

        if allow_creation:
            (
                created,
                model_version_response,
            ) = super()._get_or_create_model_version(
                model_version_request, producer_run_id=producer_run_id
            )
        else:
            try:
                model_version_response = self._get_model_version(
                    model_id=model_version_request.model,
                    version_name=model_version_request.name,
                    producer_run_id=producer_run_id,
                )
                created = False
            except KeyError:
                # The model version does not exist. We now raise the error that
                # explains why the version could not be created, instead of just
                # the KeyError that it doesn't exist
                assert error
                raise error from None

        return created, model_version_response



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Entrypoint configuration for ZenML Sagemaker pipeline steps."""

from zenml.entrypoints.step_entrypoint_configuration import (
    StepEntrypointConfiguration,
)
from zenml.pipelines.dynamic.entrypoint_configuration import (
    DynamicPipelineEntrypointConfiguration,
)
from zenml.utils.env_utils import reconstruct_environment_variables

SAGEMAKER_PROCESSOR_STEP_ENV_VAR_SIZE_LIMIT = 256


class SagemakerEntrypointConfiguration(StepEntrypointConfiguration):
    """Entrypoint configuration for ZenML Sagemaker pipeline steps.

    The only purpose of this entrypoint configuration is to reconstruct the
    environment variables that exceed the maximum length of 256 characters
    allowed for Sagemaker Processor steps from their individual components.
    """

    def run(self) -> None:
        """Runs the step."""
        # Reconstruct the environment variables that exceed the maximum length
        # of 256 characters from their individual chunks
        reconstruct_environment_variables()

        # Run the step
        super().run()


class SagemakerDynamicPipelineEntrypointConfiguration(
    DynamicPipelineEntrypointConfiguration
):
    """Entrypoint configuration for ZenML Sagemaker dynamic pipelines.

    The only purpose of this entrypoint configuration is to reconstruct the
    environment variables that exceed the maximum length of 256 characters
    allowed for Sagemaker Processor steps from their individual components.
    """

    def run(self) -> None:
        """Runs the step."""
        # Reconstruct the environment variables that exceed the maximum length
        # of 256 characters from their individual chunks
        reconstruct_environment_variables()

        # Run the step
        super().run()



"""Use Text type [7280c14811d6].

Revision ID: 7280c14811d6
Revises: 5994f9ad0489
Create Date: 2022-11-09 16:29:37.025589

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "7280c14811d6"
down_revision = "5994f9ad0489"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pipelinerunschema", schema=None) as batch_op:
        batch_op.alter_column(
            "pipeline_configuration",
            existing_type=sa.VARCHAR(length=4096),
            type_=sa.TEXT(),
            existing_nullable=False,
        )

    with op.batch_alter_table("pipelineschema", schema=None) as batch_op:
        batch_op.alter_column(
            "docstring",
            existing_type=sa.VARCHAR(length=4096),
            type_=sa.TEXT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "spec",
            existing_type=sa.VARCHAR(length=4096),
            type_=sa.TEXT(),
            existing_nullable=False,
        )

    with op.batch_alter_table("steprunschema", schema=None) as batch_op:
        batch_op.alter_column(
            "parameters",
            existing_type=sa.VARCHAR(length=4096),
            type_=sa.TEXT(),
            existing_nullable=False,
        )
        batch_op.alter_column(
            "step_configuration",
            existing_type=sa.VARCHAR(length=4096),
            type_=sa.TEXT(),
            existing_nullable=False,
        )
        batch_op.alter_column(
            "docstring",
            existing_type=sa.VARCHAR(length=4096),
            type_=sa.TEXT(),
            existing_nullable=True,
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("steprunschema", schema=None) as batch_op:
        batch_op.alter_column(
            "docstring",
            existing_type=sa.TEXT(),
            type_=sa.VARCHAR(length=4096),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "step_configuration",
            existing_type=sa.TEXT(),
            type_=sa.VARCHAR(length=4096),
            existing_nullable=False,
        )
        batch_op.alter_column(
            "parameters",
            existing_type=sa.TEXT(),
            type_=sa.VARCHAR(length=4096),
            existing_nullable=False,
        )

    with op.batch_alter_table("pipelineschema", schema=None) as batch_op:
        batch_op.alter_column(
            "spec",
            existing_type=sa.TEXT(),
            type_=sa.VARCHAR(length=4096),
            existing_nullable=False,
        )
        batch_op.alter_column(
            "docstring",
            existing_type=sa.TEXT(),
            type_=sa.VARCHAR(length=4096),
            existing_nullable=True,
        )

    with op.batch_alter_table("pipelinerunschema", schema=None) as batch_op:
        batch_op.alter_column(
            "pipeline_configuration",
            existing_type=sa.TEXT(),
            type_=sa.VARCHAR(length=4096),
            existing_nullable=False,
        )

    # ### end Alembic commands ###



"""Release [0.83.1].

Revision ID: 0.83.1
Revises: 3d7e39f3ac92
Create Date: 2025-06-23 07:26:40.361406

"""

# revision identifiers, used by Alembic.
revision = "0.83.1"
down_revision = "3d7e39f3ac92"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
